local TIMETHRESHOLD = 1200
local RAIDROSTER_DELAY = 2
local _, _, ICON_MINUS = GetSpellInfo(39092)
local _, _, ICON_PLUS = GetSpellInfo(39089)
local _, _, ICON_PRIO = GetSpellInfo(1130)
local LOOT_SELF_REGEX = gsub(LOOT_ITEM_SELF, "%%s", "(.+)")
local LOOT_REGEX = gsub(LOOT_ITEM, "%%s", "(.+)")

AsmoRaidTracker = LibStub("AceAddon-3.0"):NewAddon("AsmodeusRaidTracker", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceBucket-3.0")

--------------------------------------------------------------------------------
-- Helpers

-- Debug function
function AsmoRaidTracker:Debug(...)
	if self.db.profile.debug then
	    self:Print(...)
	end
end

-- aliases
local tinsert, tmaxn, tsort, tremove, tconcat = table.insert, table.maxn, table.sort, table.remove, table.concat
local fmt, gmatch = string.format, string.gmatch
local date = date

-- allows to iterate a table based on key alphabetical order
local function pairsByKey (t, f)
	local a = {}
	for n in pairs(t) do table.insert(a, n) end
	table.sort(a, f)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], t[a[i]]
		end
	end
	return iter
end

-- allows descending sort
local function descSorter(a, b)
	return a > b
end

-- Raid iterator
local function RaidIterator()
	local i = 0
	return function ()
		i = i + 1
		local unit
		if UnitInRaid("player") and i<=GetNumRaidMembers() then
			return "raid"..i
		else
			numPartyMembers = GetNumPartyMembers()
			if i<= numPartyMembers then
				return "party"..i
			elseif i==numPartyMembers+1 then
				return "player"
			end
		end
	end
end

--------------------------------------------------------------------------------
-- Standard Addon methods

function AsmoRaidTracker:OnInitialize()
    -- Called when the addon is loaded
    LibStub("AceConfig-3.0"):RegisterOptionsTable("AsmodeusRaidTracker", self.options, {"art"})
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("AsmodeusRaidTracker", "AsmodeusRaidTracker")
    
    self.db = LibStub("AceDB-3.0"):New("AsmodeusRaidTrackerDB", self.defaultDb, "Default")
end

function AsmoRaidTracker:OnEnable()
	-- Initializes addon
	self.isInRaid = false
	self.displayState = {
		["raidTab"] = {},
		["lootTab"] = {},
		["itemTab"] = {},
	}
	self.displayWidgets = {
		["raidTab"] = {},
		["lootTab"] = {},
		["itemTab"] = {},
	}
	self.commLog = {
		last = 0,
		first = 1,
	}
	
	-- Set up modules
    for name, module in self:IterateModules() do
    	self:EnableModule(name)
    end
    
    ----------------------------------------------------------------------------
    -- Database maintenance scripts
    
    -- maintain items length 
    local items=self.db.profile.items
    local length = 0
    for itemId, item in pairs(items) do
    	if itemId ~= "length" then
    		length = length + 1
    	end
	end
	items.length = length
	
	-- delete oldest raids
	local raids = self.db.profile.raids
	for raidKey, raid in pairs(raids) do
		if raid.beginningTimestamp<time()-32*24*3600 then
			raids[raidKey]=nil
		end
	end
end

function AsmoRaidTracker:OnDisable()
    -- Called when the addon is disabled
end

--------------------------------------------------------------------------------
-- Raid management

function AsmoRaidTracker:InitRaid(instanceName)-- SYNCHRONIZED BEWARE
	local timestamp = time()
	
	self:DbCreateRaid(instanceName, timestamp)
	local raid = self.db.profile.raids[timestamp]
	self.currentRaid = raid
	self.currentRaid_MembersStatus = {}
	
	-- populate the raid with members already in
	for unit in RaidIterator() do
		if UnitExists(unit) then
			self:OnMemberStatusChanged(UnitName(unit), "JOINED")
	    end
	end
	
	-- Register raid events
	self:RegisterRaidEvents()
	
	-- Send raid begin notification to other officers
	self.synchronizer:BroadcastRaidBegin(UnitName("player"), raid, self.currentRaid_MembersStatus)
end

function AsmoRaidTracker:Remote_InitRaid(raid, raidersStatus)
	self.db.profile.raids[raid.beginningTimestamp] = raid
	self.currentRaid = raid
	self.currentRaid_MembersStatus = raidersStatus
	self:RegisterRaidEvents()
end

function AsmoRaidTracker:RestartRecordingRaid(raidKey)
	local timestamp = time()
	
	local raid = self.db.profile.raids[raidKey]
	self.currentRaid = raid
	self.currentRaid_MembersStatus = {}
	
	-- populate the raid with members already in
	for unit in RaidIterator() do
		if UnitExists(unit) then
			self:OnMemberStatusChanged(UnitName(unit), "JOINED")
	    end
	end
	
	-- Register raid events
	self:RegisterRaidEvents()
	
	-- Send raid begin notification to other officers
	self.synchronizer:BroadcastRaidBegin(UnitName("player"), raid, self.currentRaid_MembersStatus)
end

function AsmoRaidTracker:RegisterRaidEvents()
	self.raidRosterUpdateBucketEvent = self:RegisterBucketEvent("PARTY_MEMBERS_CHANGED", RAIDROSTER_DELAY, "OnRosterUpdate")
	self:RegisterEvent("CHAT_MSG_LOOT", "OnLootMessage")
end

function AsmoRaidTracker:EndRaid()-- SYNCHRONIZED BEWARE
	local timestamp = time()
	local raid = self.currentRaid
	
	raid.endTimestamp = timestamp
	
	self.currentRaid = nil
	self.currentRaid_MembersStatus = nil
	
	-- Unregister raid events
	self:UnregisterBucket(self.raidRosterUpdateBucketEvent)
	self:UnregisterEvent("CHAT_MSG_LOOT")
	
	-- Send raid end notification
	self.synchronizer:BroadcastRaidEnd(timestamp)
end

function AsmoRaidTracker:Remote_EndRaid(timestamp)
	local raid = self.currentRaid
	
	raid.endTimestamp = timestamp
	
	self.currentRaid = nil
	self.currentRaid_MembersStatus = nil
	
	-- Unregister raid events
	self:UnregisterBucket(self.raidRosterUpdateBucketEvent)
	self:UnregisterEvent("CHAT_MSG_LOOT")
end

function AsmoRaidTracker:OnRosterUpdate()
	if not self.currentRaid then return end
	
	local raid = self.currentRaid
	local raidUnitsByName = {}
	
	-- check for joined members
	for unit in RaidIterator() do
	    if UnitExists(unit) then
			local name = UnitName(unit)
			raidUnitsByName[name] = unit
			if not raid.members[name] or (raid.members[name] and self.currentRaid_MembersStatus[name]=="OUT") then
			    self:OnMemberStatusChanged(name, "JOINED")
			end
	    end
	end
	-- check for left members
	for name, member in pairs(raid.members) do
		local unit = raidUnitsByName[name]
		if not unit and self.currentRaid_MembersStatus[name]=="IN" then
		    self:OnMemberStatusChanged(name, "LEFT")
		end
	end
end

function AsmoRaidTracker:OnMemberStatusChanged(name, action)
	local raid = self.currentRaid
	
	if not raid.members[name] then
		self:DbAddRaidMemberToRaid(raid.beginningTimestamp, name, nil, false)
		self:UpdateDisplay_RaidRoster(raid.beginningTimestamp)
	end
	
	self:DbAddRaiderLogEntry(raid.beginningTimestamp, name, action)
	
	self.currentRaid_MembersStatus[name] = (action=="JOINED") and "IN" or "OUT"

	self.synchronizer.needSync = true
end

--------------------------------------------------------------------------------
-- Loot Management

function AsmoRaidTracker:OnLootMessage(event, msg)
    local _, _, name, itemlink = string.find(msg, LOOT_REGEX)
	if not name then 
		_, _, itemlink = string.find(msg, LOOT_SELF_REGEX)
		name = UnitName("player")
	end
	if not self.currentRaid.members[name] then return end	-- Do not record loots for non raid members
	if not itemlink then return end
	local _, _, itemId = string.find(itemlink, "item:(%d+):")
	local itemName, _, itemRarity, itemLevel, _, _, _, _, _, itemTexture = GetItemInfo(itemId)
	
	-- Fix to have emblems counted with their real iLevel (DO NOT FORGET TO MANUALLY UPDATE THEM IN WTF ITEM DATABASE)
	local item = self.db.profile.items[itemId]
	if item then
		itemLevel = item.itemLevel
	end
	if not itemLevel then itemLevel = 0 end
	
	local dkpValue = 80
	local evalDkpValue, errMsg = loadstring(fmt("local itemId, itemLevel = ...;local dkpValue = 0;%s;return dkpValue", self.db.profile.evalDkpValueString), "itemEvaluationLoadstring")
	if evalDkpValue then
		dkpValue = evalDkpValue(itemId, itemLevel)
	else
		self:Print(fmt("Erreur d'évaluation du DKP : %s", errMsg))
	end
	
	if self.db.profile.itemBlacklist[itemId+0] then return end
	
	-- Do not track items if not at least epic-level
	if itemRarity < self.db.profile.itemRarityThreshold then return end
	
	local raid = self.currentRaid
	local raidKey = raid.beginningTimestamp
	local lootKey = self:DbAddLoot(raidKey, time(), name, itemId, itemName, itemTexture, itemRarity, itemLevel, dkpValue, GetLocale())
	local loot = raid.loots[lootKey]
	local item = self.db.profile.items[loot.itemId]
	self:UpdateDisplay_LootList(raidKey)
	-- Broadcast loot (only if not master)
	self.synchronizer:NotifyLootToMaster(loot, item)
	
	self.synchronizer.needSync = true
end

--------------------------------------------------------------------------------
-- Output message log

function AsmoRaidTracker:OutputCommLog()
	self:Print("Comm log : ----------------")
	local log = self.commLog
	for i = log.last, log.first, -1 do
		local comm = log[i]
		self:Print(fmt("[%s] %s@%s %s(%s)",
			date("%H:%M:%S", comm.timestamp),
			(comm.type == "IN") and "<<<" or ">>>",
			(comm.actor=="") and "Nobody" or comm.actor,
			comm.method,
			tconcat(comm.args, ", ")
		))
		if comm.error then
			self:Print(fmt("   |cFFFF0000Error |cFFFFFF[%s]", comm.error))
		end
	end
end

--------------------------------------------------------------------------------
-- Database queries

-- Config methods
function AsmoRaidTracker:DbUpdateConfig(config)
	-- copy in self.db.profile the values in config
	for k, v in pairs(config) do
		self.db.profile[k] = v
	end
end

-- Raid methods
function AsmoRaidTracker:DbGetRaidListForDropdown()
	local raidList = {}
	local firstRaidKey 
	for key, raid in pairsByKey(self.db.profile.raids, descSorter) do
		raidList[key] = date("%d/%m [%a]", raid.beginningTimestamp) .. " - " .. raid.instanceName
		if not firstRaidKey then firstRaidKey = key end
	end
	return raidList, firstRaidKey
end

-- Roster methods
function AsmoRaidTracker:DbGetMemberListForTree(raidKey)
	local memberList = {}
	local raid = self.db.profile.raids[raidKey]
	
	for _, member in pairsByKey(raid.members) do
		-----------------------------------
		-- Determine icon to display if any
		local icon
		-- Get first and last log entry for member
		local firstEntry, lastEntry = self:DbGetLogEdgeEntries(raidKey, member.name)
		if member.isPrio then
			icon = ICON_PRIO
		elseif lastEntry and raid.endTimestamp and lastEntry.timestamp < raid.endTimestamp - TIMETHRESHOLD and lastEntry.action == "LEFT" then
			icon = ICON_MINUS
		elseif firstEntry and firstEntry.action == "JOINED" and firstEntry.timestamp > raid.beginningTimestamp + TIMETHRESHOLD then
			icon = ICON_PLUS
		else
			icon = ""
		end
		
		-----------------------------------
		-- Create and insert leaf
		local leaf = {
			value = member.name,
			text = member.name,
			icon = icon,
		}
		tinsert(memberList, leaf)
	end
	return memberList
end
function AsmoRaidTracker:DbGetEligibleGuildMembers(raidKey, offlineIncluded)
	local raid = self.db.profile.raids[raidKey]
	-- construct list
	local eligibles = {}
	local index = 1	
	local name, _, rankIndex, _, _, _, _, _, online = GetGuildRosterInfo(index) 
	while name do
		-- If offline are included, we consider everyone as online
		if offlineIncluded then online = true end
		-- If member is online, not in raid and not part of an excluded rank		
		if online and not raid.members[name] and not self.db.profile.excludedGuildRanks[rankIndex] then 		
			eligibles[name] = name
		end
		index = index + 1
		name, rank, rankIndex, level, _, _, _, _, online = GetGuildRosterInfo(index)		
	end
	
	return eligibles
end
function AsmoRaidTracker:DbGetRaidersNumber(raidKey)
	local raid = self.db.profile.raids[raidKey]    	
	local numTotal, numPrio = 0, 0
	for name, member in pairs(raid.members) do
		numTotal = numTotal + 1
		if member.isPrio then
			numPrio = numPrio + 1
		end
	end
	return numTotal, numPrio
end

-- Raider log methods
function AsmoRaidTracker:DbGetLogEdgeEntries(raidKey, raiderName)
	local raid = self.db.profile.raids[raidKey]
	local log = raid.members[raiderName].log
	local firstEntry, lastEntry
	for entryTimestamp, entry in pairs(log) do
		if not lastEntry or entry.timestamp>lastEntry.timestamp then
			lastEntry = entry
		end
		if not firstEntry or entry.timestamp<firstEntry.timestamp then
			firstEntry = entry
		end
	end
	return firstEntry, lastEntry
end

--------------------------------------------------------------------------------
-- Database updates

-- Raid methods
function AsmoRaidTracker:DbDeleteRaid(timestamp)
	self.db.profile.raids[timestamp] = nil
end
function AsmoRaidTracker:DbCreateRaid(instanceName, timestamp)
	local raid = {
	    instanceName = instanceName,
	    beginningTimestamp = timestamp,
	    members = {},
	    loots = {},
	}
	self.db.profile.raids[timestamp] = raid
end

-- Raid members methods
function AsmoRaidTracker:DbRemoveRaiderFromRaid(selectedRaid, selectedMember)
	local raid = self.db.profile.raids[selectedRaid]
	raid.members[selectedMember] = nil
end
function AsmoRaidTracker:DbAddRaidMemberToRaid(raidKey, name, isPrio)
	local raid = self.db.profile.raids[raidKey]
	local timestamp = time()
	
	local member = {
        name = name,
        log = {},
        isPrio = isPrio
	}
	
	raid.members[member.name] = member
end
function AsmoRaidTracker:DbAddRaiderLogEntry(raidKey, name, action)
	local raid = self.db.profile.raids[raidKey]
	local log = raid.members[name].log
	local timestamp = time()
	
	local entry = {
		timestamp = timestamp,
		action = action
	}
	
	log[timestamp] = entry
end
function AsmoRaidTracker:DbSetRaiderPrio(selectedRaid, selectedRaider, isPrio)
	local raid = self.db.profile.raids[selectedRaid]
	local member = raid.members[selectedRaider]
	member.isPrio = isPrio
end

-- Roster methods
function AsmoRaidTracker:DbAddRaiderToRoster(raiderName)
	local raider = {
		name = raiderName,
		isTank = false,
		isHeal = false,
		isMelee = false,
		isRanged = false,
	}
	self.db.profile.raiders[raider.name] = raider
end

-- Loot methods
function AsmoRaidTracker:DbAddLoot(raidKey, timeStamp, looterName, itemId, itemName, itemTexture, itemRarity, itemLevel, dkpValue, locale)
	local raid = self.db.profile.raids[raidKey]
	local loots = raid.loots 
	
	-- Compute key
	local index = 1
	while loots[timeStamp.."-"..index] do
		index = index + 1
	end
	local key = timeStamp.."-"..index
	
	-- create loot instance
	local loot = {
		timestamp = timeStamp,
		looterName = looterName,
		originalLooterName = looterName, 
		itemId = itemId,
		dkpValue = dkpValue,
	}
	
	if not self.db.profile.items[itemId] then
		self:DbAddItem(itemId, itemName, itemTexture, itemRarity, itemLevel, locale)
	end
	
	-- insert into loot database
	loots[key] = loot
	
	return key
end

function AsmoRaidTracker:DbChangeLooter(raidKey, lootKey, looterName)
	local raid = self.db.profile.raids[raidKey]
	local loot = raid.loots[lootKey] 
	loot.looterName = looterName
end

-- Item methods
function AsmoRaidTracker:DbAddItem(itemId, itemName, itemTexture, itemRarity, itemLevel, locale)
	-- create item instance
	local item = {
		itemId = itemId,
		itemName = itemName,
		itemTexture = itemTexture,
		itemRarity = itemRarity,
		itemLevel = itemLevel,
		locale = locale,
	}
	
	-- insert into item database
	if not self.db.profile.items[itemId] then
		self.db.profile.items.length = self.db.profile.items.length + 1
	end
	self.db.profile.items[itemId] = item
end
function AsmoRaidTracker:DbRemoveItem(itemId)
	self.db.profile.items[itemId] = nil
	self.db.profile.items.length = self.db.profile.items.length-1
end
function AsmoRaidTracker:DbRemoveUnusedItems()
	local usedItems = {}
	for _, raid in pairs(self.db.profile.raids) do
		for _, loot in pairs(raid.loots) do
			usedItems[loot.itemId] = true
		end
	end
	
	for itemId, _ in pairs(self.db.profile.items) do
		if not (usedItems[itemId] or itemId == "length") then
			self:DbRemoveItem(itemId)
		end
	end
end


-- Comm log management
function AsmoRaidTracker:DbLogComm(comm)
	comm.timestamp = time()
    self.commLog.last = self.commLog.last+1
	self.commLog[self.commLog.last] = comm
	if self.commLog.last - self.commLog.first > 10 then
	    self.commLog[self.commLog.first] = nil
	    self.commLog.first = self.commLog.first + 1  
	end
end