--------------------------------------------------------------------------------
-- Speeding up global recovery
local AsmoRaidTracker = AsmoRaidTracker

--------------------------------------------------------------------------------
-- Helpers

-- Aliases
local date = date
local fmt, sfind, slower = string.format, string.find, string.lower
local tremove, tinsert, tconcat = table.remove, table.insert, table.concat

-- allows descending sort
local function descSorter(a, b)
	return a >= b
end

-- allows to iterate a table based on key alphabetical order
local function pairsByKey (t, f)
	local a = {}
	for n in pairs(t) do table.insert(a, n) end
	table.sort(a, f)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], t[a[i]]
		end
	end
	return iter
end

-- allows to iterate loots
local function LootFilter (t, f, itemNameFilter, looterNameFilter, disenchantFlag, freeFlag, LQEFlag)
	local a = {}
	local items = AsmoRaidTracker.db.profile.items
	itemNameFilter = slower(itemNameFilter)
	looterNameFilter = slower(looterNameFilter)
	for n in pairs(t) do
		local item = items[t[n].itemId]
		-- item non trouvé, on renvoie l'entrée
		if not item then
			table.insert(a, n)
		-- L'item n'est ni free, ni dez, ni lqe, il doit satisfaire aux filtres
		elseif
			t[n].looterName ~= "*DISENCHANT" and
			t[n].looterName ~= "*FREE" and
			t[n].looterName ~= "*LQE" and 
			sfind(slower(t[n].looterName), looterNameFilter) and 
			sfind(item.itemName, itemNameFilter)
		then
			table.insert(a, n)
		-- sinon on vérifie les flags et le nom
		elseif
			(
				(t[n].looterName == "*DISENCHANT" and disenchantFlag) or
				(t[n].looterName == "*FREE" and freeFlag) or
				(t[n].looterName == "*LQE" and LQEFlag)
			) and sfind(item.itemName, itemNameFilter)
		then
			table.insert(a, n)
		end 
	end
	table.sort(a, f)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], t[a[i]]
		end
	end
	return iter
end

-- allows to iterate items
local function ItemFilter (t, f, itemNameFilter)
	local a = {}
	itemNameFilter = slower(itemNameFilter)
	for n in pairs(t) do
		if n~="length"  and sfind(slower(t[n].itemName), itemNameFilter)	then
			table.insert(a, n)
		end 
	end
	table.sort(a, f)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], t[a[i]]
		end
	end
	return iter
end

--------------------------------------------------------------------------------
-- Libraries instanciation
local GUI = LibStub("AceGUI-3.0")
local LibQTip = LibStub('LibQTip-1.0')

--------------------------------------------------------------------------------
-- Constants
local RAIDWINDOW_TABLIST = {
	{text="Raid", value="raid"},
	{text="Loots", value="loots"},
	{text="Items", value="items"},
	{text="Export", value="export"},
}
local RARITY_COLORS = {
	[0] = "9D9D9D",
	[1] = "FFFFFF",
	[2] = "1EFF00",
	[3] = "0070FF",
	[4] = "A335EE",
	[5] = "FF8000",
	[6] = "E6CC80",
	[7] = "E6CC80",
}
local _, _, ICON_LOG = GetSpellInfo(31460)

--------------------------------------------------------------------------------
-- Static popups

-- Popup for raid confirmation
StaticPopupDialogs["ASMORAIDTRACKER_CONFIRMRAID"] = {
	text = "Comencer un raid de ce nom ?",
	button1 = "OK",
	button2 = "Annuler",
	OnAccept = function(self)
		local instanceName = self.editBox:GetText()
		AsmoRaidTracker:InitRaid(instanceName)
		AsmoRaidTracker:UpdateDisplay_SelectedRaid("ASMORAIDTRACKER_BEGINRAID")
	end,
	OnShow = function (self, data)
		self.editBox:SetText(GetRealZoneText())
	end,
	timeout = 0,
	whileDead = true,
	hasEditBox = true,
	hideOnEscape = true,
}

--------------------------------------------------------------------------------
-- Custom layout for loot list

-- bunch of things to define safecall function
local tconcat = table.concat
local xpcall = xpcall
local function errorhandler(err)
	return geterrorhandler()(err)
end
local function CreateDispatcher(argCount)
	local code = [[
		local xpcall, eh = ...
		local method, ARGS
		local function call() return method(ARGS) end
	
		local function dispatch(func, ...)
			method = func
			if not method then return end
			ARGS = ...
			return xpcall(call, eh)
		end
	
		return dispatch
	]]
	
	local ARGS = {}
	for i = 1, argCount do ARGS[i] = "arg"..i end
	code = code:gsub("ARGS", tconcat(ARGS, ", "))
	return assert(loadstring(code, "safecall Dispatcher["..argCount.."]"))(xpcall, errorhandler)
end
local Dispatchers = setmetatable({}, {__index=function(self, argCount)
	local dispatcher = CreateDispatcher(argCount)
	rawset(self, argCount, dispatcher)
	return dispatcher
end})
Dispatchers[0] = function(func)
	return xpcall(func, errorhandler)
end
local function safecall(func, ...)
	return Dispatchers[select('#', ...)](func, ...)
end

GUI:RegisterLayout("AsmoRaidTracker_LootRow",
	function(content, children)
		local height = content.height or content:GetHeight() or 0
		local width = 0
		for i = 1, #children do
			local child = children[i]
			
			local frame = child.frame
			frame:ClearAllPoints()
			frame:Show()
			if i == 1 then
				frame:SetPoint("TOPLEFT",content,"TOPLEFT",0,-10)
			elseif i==3 then
			    if child.type=="Dropdown" then
					frame:SetPoint("TOPLEFT",children[i-1].frame,"TOPRIGHT",0,28)
				else
				    frame:SetPoint("TOPLEFT",children[i-1].frame,"TOPRIGHT",0,0)
				end
			else
				frame:SetPoint("TOPLEFT",children[i-1].frame,"TOPRIGHT",0,0)
			end
			
			width = width + (frame.width or frame:GetWidth() or 0)
		end
		safecall( content.obj.LayoutFinished, content.obj, nil, 25 )
	end
)

GUI:RegisterLayout("AsmoRaidTracker_ItemRow",
	function(content, children)
		local height = content.height or content:GetHeight() or 0
		local width = 0
		for i = 1, #children do
			local child = children[i]
			
			local frame = child.frame
			frame:ClearAllPoints()
			frame:Show()
			if i == 1 then
				frame:SetPoint("TOPLEFT",content,"TOPLEFT",0,0)
			elseif i == 2 then
				frame:SetPoint("TOPLEFT",children[i-1].frame,"TOPRIGHT",0,6)
			else
				frame:SetPoint("TOPLEFT",children[i-1].frame,"TOPRIGHT",0,0)
			end
			
			width = width + (frame.width or frame:GetWidth() or 0)
		end
		safecall( content.obj.LayoutFinished, content.obj, nil, 25 )
	end
)

--------------------------------------------------------------------------------
-- Callbacks declaration
local Window_RaidTab_Tree_OnSelectRaider
local Window_RaidTab_AddRaiderChangePrio
local Window_RaidTab_AddRaiderSelectGuildMember
local Window_RaidTab_AddRaiderIncludeOffline
local Window_RaidTab_AddRaider
local Window_RaidTab_RemoveRaider
local Window_BeginRaid
local Window_RestartRecordingRaid
local Window_EndRaid
local Window_DeleteRaid
local Window_SendRaid
local Window_RaidFilter
local Window_OnSelectTab
local Window_OnClose
local Window_AssumeControlRaid
local Window_RaidTab_RaiderSetPrio
local Window_RaidTab_LogRaider_OnEnter
local Window_RaidTab_LogRaider_OnLeave
local Window_LootTab_FilterChanged
local Window_LootTab_FilterItemName
local Window_LootTab_FilterLooterName
local Window_LootTab_FilterDisenchantFlag
local Window_LootTab_FilterFreeFlag
local Window_LootTab_FilterLQEFlag
local Window_LootTab_ItemNameEnter
local Window_LootTab_ItemNameLeave
local Window_LootTab_LootList_SelectLooter
local Window_ItemTab_FilterChanged
local Window_ItemTab_FilterItemName
local Window_ItemTab_RemoveItem
local Window_ItemTab_BlacklistItem
local Window_ItemTab_RemoveUnusedItems

--------------------------------------------------------------------------------
-- Callbacks for raid tab

-- callback for select raid member
Window_RaidTab_Tree_OnSelectRaider = function(treeGroup, event, raiderName)
	local self = AsmoRaidTracker
	local selectedRaid = self.displayState.selectedRaid
	
	self.displayState.raidTab.selectedRaider = raiderName
	
	local tabGroup = self.displayWidgets.tabGroup
	tabGroup:ReleaseChildren()
	self:DrawRaidTab(tabGroup, selectedRaid)
end

-- Callback for add raider
Window_RaidTab_AddRaiderChangePrio = function(widget, event, v)
	local self = AsmoRaidTracker
	self.displayState.raidTab.addRaiderPrio = v
end
Window_RaidTab_AddRaiderSelectGuildMember = function(widget, event, v)
	local self = AsmoRaidTracker
	self.displayState.raidTab.addRaiderGuildMember = v
end
Window_RaidTab_AddRaiderIncludeOffline = function(widget, event, v)
	local self = AsmoRaidTracker
	local raid = self.db.profile.raids[self.displayState.selectedRaid]
	self.displayState.raidTab.addRaiderIncludeOffline = v
	
	self.displayWidgets.raidTab.addRaiderSelectMember:SetList(self:DbGetEligibleGuildMembers(raid.beginningTimestamp, v))
end
Window_RaidTab_AddRaider = function(widget, event)
	local self = AsmoRaidTracker
	local selectedRaid = self.displayState.selectedRaid
	local memberName = self.displayState.raidTab.addRaiderGuildMember
	local addRaiderPrio = self.displayState.raidTab.addRaiderPrio
	
	if self.displayState.raidTab.addRaiderGuildMember then
		self:DbAddRaidMemberToRaid(selectedRaid, memberName, addRaiderPrio)
		self:DbAddRaiderLogEntry(selectedRaid, memberName, "ADDED")
		self.displayState.raidTab.addRaiderGuildMember = nil
	end
	Window_RaidTab_Tree_OnSelectRaider(self.displayWidgets.raidTab.memberTreeGroup, "ASMORAIDTRACKER_ADDEDRAIDER_SELECTRAIDER", memberName)
	
	AsmoRaidTracker.synchronizer.needSync = true
end

-- callback for remove raider button
Window_RaidTab_RemoveRaider = function()
	local self = AsmoRaidTracker
	local selectedRaid = self.displayState.selectedRaid
	local selectedRaider = self.displayState.raidTab.selectedRaider

    self:DbRemoveRaiderFromRaid(selectedRaid, selectedRaider)
	
	Window_OnSelectTab(self.displayWidgets.tabGroup, "ASMORAIDTRACKER_REMOVERAIDER", "raid")
	
	AsmoRaidTracker.synchronizer.needSync = true
end

-- callback for set raider prio button
Window_RaidTab_RaiderSetPrio = function(widget, event, v)
	local self = AsmoRaidTracker
	local selectedRaid = self.displayState.selectedRaid
	local selectedRaider = self.displayState.raidTab.selectedRaider

	self:DbSetRaiderPrio(selectedRaid, selectedRaider, v)
	
	Window_RaidTab_Tree_OnSelectRaider(self.displayWidgets.raidTab.memberTreeGroup, "ASMORAIDTRACKER_RAIDERSETPRIO_SELECTRAIDER", selectedRaider)
	
	AsmoRaidTracker.synchronizer.needSync = true
end

-- Log tooltip display callbacks
Window_RaidTab_LogRaider_OnEnter = function(widget, event)
	local self = AsmoRaidTracker
	
	-- Acquire tooltip
	local tooltip = LibQTip:Acquire("LogTooltip", 3, "LEFT", "CENTER", "RIGHT")
	self.displayWidgets.raidTab.tooltip = tooltip 
	
	-- Add header
	tooltip:AddHeader("Heure", "", "Action")
	-- For each entry in log : display line
	local raid = self.db.profile.raids[self.displayState.selectedRaid]
	local log = raid.members[self.displayState.raidTab.selectedRaider].log
	if log then
		for _, entry in pairsByKey(log) do
			tooltip:AddLine(date("%H:%M", entry.timestamp),">", entry.action)
		end
	end
	
	-- Use smart anchoring code to anchor the tooltip to our frame
	tooltip:SmartAnchorTo(widget.frame)
	
	-- Show it, et voilà !
	tooltip:Show()
end
Window_RaidTab_LogRaider_OnLeave = function(widget, event)
	local self = AsmoRaidTracker
	
	tooltip = self.displayWidgets.raidTab.tooltip
	LibQTip:Release(tooltip)
   	self.displayWidgets.raidTab.tooltip = nil
end


--------------------------------------------------------------------------------
-- Drawing raid tab

function AsmoRaidTracker:DrawRaidTab(tabGroup, raidKey)
	tabGroup:SetLayout("Fill")
	
	local memberList = self:DbGetMemberListForTree(raidKey)
	local raid = self.db.profile.raids[raidKey]
	local selectedRaider = self.displayState.raidTab.selectedRaider
	
	local tg = GUI:Create("TreeGroup")
	self.displayWidgets.raidTab.memberTreeGroup = tg
	tg:SetLayout("List")
	tg:SetTree(memberList)
	if selectedRaider and raid.members[selectedRaider] then
		tg:SelectByValue(self.displayState.raidTab.selectedRaider)
	end
	tg:SetCallback("OnGroupSelected", Window_RaidTab_Tree_OnSelectRaider)
	tg:EnableButtonTooltips(false)
	tabGroup:AddChild(tg)
	
	----------------------------------
	-- Raid
		
	-- Raid inline group
	local igRaid = GUI:Create("InlineGroup")
	igRaid:SetTitle("Raid")
	igRaid:SetFullWidth(true)
	igRaid:SetLayout("Flow")
	tg:AddChild(igRaid)
	
	-- Raid info 1
	local lRaidInfo = GUI:Create("Label")
	lRaidInfo:SetFullWidth(true)
	lRaidInfo:SetText(fmt("%s (%s) de %s à %s",
		raid.instanceName, 
		date("%Y/%m/%d", raid.beginningTimestamp), 
		date("%H:%M", raid.beginningTimestamp),
		raid.endTimestamp and date("%H:%M", raid.endTimestamp) or "[En cours]"
	))
	igRaid:AddChild(lRaidInfo)
	
	-- Raid info 2
	local lRaidInfo2 = GUI:Create("Label")
	lRaidInfo2:SetFullWidth(true)
	lRaidInfo2:SetText(fmt("%s présents (%s prios)", self:DbGetRaidersNumber(raid.beginningTimestamp)))
	igRaid:AddChild(lRaidInfo2)
	
	-- Raid info 3
	local lRaidInfo3 = GUI:Create("Label")
	lRaidInfo3:SetFullWidth(true)
	-- Si le raid sélectionné est le raid en cours
	if raid == self.currentRaid then
	    -- Si l'utilisateur est maître
	    if not self.synchronizer.masterName then
	    	lRaidInfo3:SetText("Ce raid est en cours (mode : Maître), possibilité de modifier les participants et attribs loots. Synchronisation avec les esclaves toutes les 10 secondes.")
		-- Sinon il est esclave
		else
		    lRaidInfo3:SetText(fmt("Ce raid est en cours (mode : Esclave), vous n'avez pas la possibilité de modifier les participants et attribs loots (cliquer sur 'Prendre contrôle' pour le faire). Synchronisation envoyée par le maître (%s) toutes les 10 secondes.", self.synchronizer.masterName))
		end
	else
	    lRaidInfo3:SetText("Ce raid est terminé, possibilité de modifier les participants et attribs loot, ne pas oublier de synchroniser avec ses petits camarades une fois les modifs faites")
	end
	igRaid:AddChild(lRaidInfo3)
	
	-- Si le raid sélectionné est le raid en cours et l'utilisateur est maître ou que le raid sélectionné n'est pas en cours
	if (raid == self.currentRaid and not self.synchronizer.masterName) or raid ~= self.currentRaid then
		-- Simple group to add raid member
		local gAddRaidMember = GUI:Create("SimpleGroup")
		gAddRaidMember:SetFullWidth(true)
		gAddRaidMember:SetLayout("Flow")
		igRaid:AddChild(gAddRaidMember)
		-- Prio checkbox
		local cbAddPrioRaidMember = GUI:Create("CheckBox")
		cbAddPrioRaidMember:SetWidth(75)
		cbAddPrioRaidMember:SetType("checkbox")
		cbAddPrioRaidMember:SetLabel("Set prio")
		cbAddPrioRaidMember:SetValue(true)
		self.displayState.raidTab.addRaiderPrio = true
		cbAddPrioRaidMember:SetTriState(false)
		cbAddPrioRaidMember:SetCallback("OnValueChanged", Window_RaidTab_AddRaiderChangePrio)
		gAddRaidMember:AddChild(cbAddPrioRaidMember)
		-- Include offline checkbox
		local cbAddRaidMemberIncludeOffline = GUI:Create("CheckBox")
		cbAddRaidMemberIncludeOffline:SetWidth(75)
		cbAddRaidMemberIncludeOffline:SetType("checkbox")
		cbAddRaidMemberIncludeOffline:SetLabel("Offline")
		cbAddRaidMemberIncludeOffline:SetValue(false)
		self.displayState.raidTab.addRaiderIncludeOffline = false
		cbAddRaidMemberIncludeOffline:SetTriState(false)
		cbAddRaidMemberIncludeOffline:SetCallback("OnValueChanged", Window_RaidTab_AddRaiderIncludeOffline)
		gAddRaidMember:AddChild(cbAddRaidMemberIncludeOffline)
		-- Guild member select
		local ddAddRaidMember = GUI:Create("Dropdown")
		self.displayWidgets.raidTab.addRaiderSelectMember = ddAddRaidMember
		ddAddRaidMember:SetWidth(150)
		ddAddRaidMember:SetList(self:DbGetEligibleGuildMembers(raid.beginningTimestamp, false))
		ddAddRaidMember:SetLabel("Ajouter membres")
		ddAddRaidMember:SetText("Sélectionner")
		self.displayState.raidTab.addRaiderGuildMember = nil
		ddAddRaidMember:SetCallback("OnValueChanged", Window_RaidTab_AddRaiderSelectGuildMember)
		gAddRaidMember:AddChild(ddAddRaidMember)
		-- OK button
		local buAddRaidMemberConfirm = GUI:Create("Button")
		buAddRaidMemberConfirm:SetWidth(50)
		buAddRaidMemberConfirm:SetText("OK")
		buAddRaidMemberConfirm:SetCallback("OnClick", Window_RaidTab_AddRaider)
		gAddRaidMember:AddChild(buAddRaidMemberConfirm)
	end
	
	-- Raider inline group
	if selectedRaider and raid.members[selectedRaider] then
		local igRaider = GUI:Create("InlineGroup")
		igRaider:SetTitle(selectedRaider)
		igRaider:SetFullWidth(true)
		igRaider:SetLayout("Flow")
		tg:AddChild(igRaider)

		local log = raid.members[selectedRaider].log
		if log then
			local buLog = GUI:Create("Button")
			buLog:SetWidth(75)
			buLog:SetText("Log")
			buLog:SetCallback("OnEnter", Window_RaidTab_LogRaider_OnEnter)
			buLog:SetCallback("OnLeave", Window_RaidTab_LogRaider_OnLeave)
			igRaider:AddChild(buLog)
		end
        -- Si le raid sélectionné est le raid en cours et l'utilisateur est maître ou que le raid sélectionné n'est pas en cours
		if (raid == self.currentRaid and not self.synchronizer.masterName) or raid ~= self.currentRaid then
	        -- button remove raider
			local buRemoveRaider = GUI:Create("Button")
			buRemoveRaider:SetWidth(150)
			buRemoveRaider:SetText("Supprimer raider")
			buRemoveRaider:SetCallback("OnClick", Window_RaidTab_RemoveRaider)
			igRaider:AddChild(buRemoveRaider)

			-- set prio checkbox
			local cbSetPrio = GUI:Create("CheckBox")
			cbSetPrio:SetWidth(75)
			cbSetPrio:SetType("checkbox")
			cbSetPrio:SetLabel("Prio")
			cbSetPrio:SetValue(raid.members[selectedRaider].isPrio)
			cbSetPrio:SetTriState(false)
			cbSetPrio:SetCallback("OnValueChanged", Window_RaidTab_RaiderSetPrio)
			igRaider:AddChild(cbSetPrio)
		end
	end
end

--------------------------------------------------------------------------------
-- Callbacks for loot tab

-- Item change mutualized callback
Window_LootTab_FilterChanged = function()
	local self = AsmoRaidTracker
	
	self.displayWidgets.lootTab.lootList:ReleaseChildren()
	AsmoRaidTracker:DrawLootList(self.displayWidgets.lootTab.lootList, self.displayState.selectedRaid)
end

-- Item name filter callback
Window_LootTab_FilterItemName = function(widget, event, v)
	local self = AsmoRaidTracker
	
	self.displayState.lootTab.filterItemName = v
	
	Window_LootTab_FilterChanged()
end

-- Looter name filter callback
Window_LootTab_FilterLooterName = function(widget, event, v)
	local self = AsmoRaidTracker
	
	self.displayState.lootTab.filterLooterName = v
	
	Window_LootTab_FilterChanged()
end

-- Disenchant flag filter callback
Window_LootTab_FilterDisenchantFlag = function(widget, event, v)
	local self = AsmoRaidTracker
	
	self.displayState.lootTab.filterDisenchantFlag = v
	
	Window_LootTab_FilterChanged()
end

-- Free flag filter callback
Window_LootTab_FilterFreeFlag = function(widget, event, v)
	local self = AsmoRaidTracker
	
	self.displayState.lootTab.filterFreeFlag = v
	
	Window_LootTab_FilterChanged()
end

-- LQE flag filter callback
Window_LootTab_FilterLQEFlag = function(widget, event, v)
	local self = AsmoRaidTracker
	
	self.displayState.lootTab.filterLQEFlag = v
	
	Window_LootTab_FilterChanged()
end

Window_LootTab_ItemNameEnter = function(widget, event)
	local itemlink = widget:GetUserData("itemlink")
	GameTooltip:SetOwner(widget.frame, "ANCHOR_LEFT ")
	GameTooltip:SetHyperlink(itemlink)
	GameTooltip:Show()
end

Window_LootTab_ItemNameLeave = function(widget, event)
	GameTooltip:Hide()
end

Window_LootTab_LootList_SelectLooter = function(widget, event, v)
	local self = AsmoRaidTracker
	
	local lootKey = widget:GetUserData("lootKey")
	local raidKey = self.displayState.selectedRaid 
	
	self:DbChangeLooter(raidKey, lootKey, v)
	
	self.displayWidgets.lootTab.lootList:ReleaseChildren()
	AsmoRaidTracker:DrawLootList(self.displayWidgets.lootTab.lootList, self.displayState.selectedRaid)
	
	AsmoRaidTracker.synchronizer.needSync = true
end


--------------------------------------------------------------------------------
-- Drawing loot tab

function AsmoRaidTracker:DrawLootList(lootList, raidKey)
	local raid = self.db.profile.raids[raidKey]
	local loots = raid.loots or {}
	local filterItemName = self.displayState.lootTab.filterItemName
	local filterLooterName = self.displayState.lootTab.filterLooterName
	local disenchantFlag = self.displayState.lootTab.filterDisenchantFlag
	local freeFlag = self.displayState.lootTab.filterFreeFlag
	local LQEFlag = self.displayState.lootTab.filterLQEFlag

	-- Creating raid member list for use in looter dropdown
	local raidMemberList = {}
	for _, member in pairsByKey(raid.members) do
		raidMemberList[member.name] = member.name
	end
	raidMemberList["*DISENCHANT"] = "Dez"
	raidMemberList["*FREE"] = "Free"
	raidMemberList["*LQE"] = "LQE"
	
	-- Create loot list
	for key, loot in LootFilter(loots, descSorter, filterItemName, filterLooterName, disenchantFlag, freeFlag, LQEFlag) do
		local item = self.db.profile.items[loot.itemId]
		
		if not item then
			item = {
				itemId = 0,
				itemName = "Item non trouvé en db",
				itemTexture = "",
				itemRarity = 7,
			}
			self.synchronizer:BroadcastForItem(loot.itemId)
		elseif not item.itemRarity then
			item.itemRarity = 5
		end
		
		
		--------------------------
		-- the overlying group 
		local sgRow = GUI:Create("SimpleGroup")
		sgRow:SetLayout("AsmoRaidTracker_LootRow")
		sgRow:SetFullWidth(true)
		sgRow:SetFullHeight(false)
		lootList:AddChild(sgRow)
		-- Hour label
		local lHour = GUI:Create("Label")
		lHour:SetWidth(75)
		lHour:SetText(date("%H:%M:%S", loot.timestamp))
		sgRow:AddChild(lHour)
		-- Item name
		local ilItemName = GUI:Create("InteractiveLabel")
		ilItemName:SetWidth(300)
		if loot.dkpValue then
			ilItemName:SetText(fmt("   |T%s:40:40:0:10|t  |cFF%s%s |cFFFFFFFF(%s DKP)",
				item.itemTexture, 
				RARITY_COLORS[item.itemRarity], 
				item.itemName,
				loot.dkpValue
			))
		else  -- TODO : delete this when all loots have dkp value
			ilItemName:SetText(fmt("   |T%s:40:40:0:10|t  |cFF%s%s",
				item.itemTexture, 
				RARITY_COLORS[item.itemRarity], 
				item.itemName
			))
		end
		local _, itemlink = GetItemInfo(item.itemId)
		if itemlink then
			ilItemName:SetUserData("itemlink", itemlink)
			ilItemName:SetCallback("OnEnter", Window_LootTab_ItemNameEnter)
			ilItemName:SetCallback("OnLeave", Window_LootTab_ItemNameLeave)
		end
		sgRow:AddChild(ilItemName)
		-- Si le raid sélectionné est le raid en cours et l'utilisateur est maître ou que le raid sélectionné n'est pas en cours
		-- On permet le changement de looter
		if (raid == self.currentRaid and not self.synchronizer.masterName) or raid ~= self.currentRaid then
			-- Looter name
			local ddLooterName = GUI:Create("Dropdown")
			ddLooterName:SetWidth(150)
			ddLooterName:SetList(raidMemberList)
			ddLooterName:SetValue(loot.looterName)
			ddLooterName:SetLabel(" ")
			ddLooterName:SetUserData("lootKey", key)
			ddLooterName:SetCallback("OnValueChanged", Window_LootTab_LootList_SelectLooter)
			sgRow:AddChild(ddLooterName)
		-- Sinon on affiche juste le nom du looter
		else
		    local lLooterName = GUI:Create("Label")
			lLooterName:SetWidth(150)
			lLooterName:SetText(loot.looterName)
			sgRow:AddChild(lLooterName)
		end
  	end
end

function AsmoRaidTracker:DrawLootsTab(tabGroup, raidKey)
	local raid = self.db.profile.raids[raidKey]
	
	tabGroup:SetLayout("Flow")

	---------------------------------------
	-- Filter box
	local igFilter = GUI:Create("InlineGroup")
	igFilter:SetTitle("Filtres")
	igFilter:SetFullWidth(true)
	igFilter:SetLayout("Flow")
	igFilter:SetHeight(10)
	tabGroup:AddChild(igFilter)
	
	-- Item name filter
	local ebFilterByItemName = GUI:Create("EditBox")
	ebFilterByItemName:SetWidth(150)
	ebFilterByItemName:SetLabel("Item")
	ebFilterByItemName:SetText("")
	self.displayState.lootTab.filterItemName = ""
	ebFilterByItemName:SetCallback("OnTextChanged", Window_LootTab_FilterItemName)
	igFilter:AddChild(ebFilterByItemName)
	-- Looter name filter
	local ebFilterByItemName = GUI:Create("EditBox")
	ebFilterByItemName:SetWidth(150)
	ebFilterByItemName:SetLabel("Looter")
	ebFilterByItemName:SetText("")
	self.displayState.lootTab.filterLooterName = ""
	ebFilterByItemName:SetCallback("OnTextChanged", Window_LootTab_FilterLooterName)
	igFilter:AddChild(ebFilterByItemName)
	-- Disenchant checkbox
	local cbFilterDisenchant = GUI:Create("CheckBox")
	cbFilterDisenchant:SetWidth(75)
	cbFilterDisenchant:SetType("checkbox")
	cbFilterDisenchant:SetLabel("Dez")
	cbFilterDisenchant:SetValue(true)
	self.displayState.lootTab.filterDisenchantFlag = true
	cbFilterDisenchant:SetTriState(false)
	cbFilterDisenchant:SetCallback("OnValueChanged", Window_LootTab_FilterDisenchantFlag)
	igFilter:AddChild(cbFilterDisenchant)
	-- Free checkbox
	local cbFilterFree = GUI:Create("CheckBox")
	cbFilterFree:SetWidth(75)
	cbFilterFree:SetType("checkbox")
	cbFilterFree:SetLabel("Free")
	cbFilterFree:SetValue(true)
	self.displayState.lootTab.filterFreeFlag = true
	cbFilterFree:SetTriState(false)
	cbFilterFree:SetCallback("OnValueChanged", Window_LootTab_FilterFreeFlag)
	igFilter:AddChild(cbFilterFree)
	-- LQE checkbox
	local cbFilterLQE = GUI:Create("CheckBox")
	cbFilterLQE:SetWidth(75)
	cbFilterLQE:SetType("checkbox")
	cbFilterLQE:SetLabel("LQE")
	cbFilterLQE:SetValue(true)
	self.displayState.lootTab.filterLQEFlag = true
	cbFilterLQE:SetTriState(false)
	cbFilterLQE:SetCallback("OnValueChanged", Window_LootTab_FilterLQEFlag)
	igFilter:AddChild(cbFilterLQE)
	
	---------------------------------------
	-- Loot list
	
	-- Scroll frame
	local scLootList = GUI:Create("ScrollFrame")
	self.displayWidgets.lootTab.lootList = scLootList
	scLootList:SetLayout("Flow")
	scLootList:SetFullWidth(true)
	scLootList:SetFullHeight(true)	
	tabGroup:AddChild(scLootList)
	
	self:DrawLootList(scLootList, raidKey) 	    
end

--------------------------------------------------------------------------------
-- Callbacks for items tab

Window_ItemTab_FilterChanged = function()
	local self = AsmoRaidTracker
	
	self.displayWidgets.itemTab.itemList:ReleaseChildren()
	AsmoRaidTracker:DrawItemList(self.displayWidgets.itemTab.itemList)
end

Window_ItemTab_FilterItemName = function(widget, event, v)
	local self = AsmoRaidTracker
	
	self.displayState.itemTab.filterItemName = v
	
	Window_ItemTab_FilterChanged()
end

Window_ItemTab_RemoveItem = function(widget, event)
	local self = AsmoRaidTracker
	
	local itemId = widget:GetUserData("itemid")	
	self:DbRemoveItem(itemId)
	
	Window_OnSelectTab(self.displayWidgets.tabGroup, "ASMORAIDTRACKER_REMOVEITEM", "items")
end

Window_ItemTab_BlacklistItem = function(widget, event)
	local self = AsmoRaidTracker

	local itemId = widget:GetUserData("itemid")
	-- remove item from database
	self:DbRemoveItem(itemId)
	-- remove all loots pertaining to this item
	for _, raid in pairs(self.db.profile.raids) do
		for lootKey, loot in pairs(raid.loots) do
		    if loot.itemId == itemId then
		        raid.loots[lootKey] = nil
		    end
		end
	end
	-- Add item to blacklist
	self.db.profile.itemBlacklist[itemId+0]=true
	-- TODO : broadcast

	Window_OnSelectTab(self.displayWidgets.tabGroup, "ASMORAIDTRACKER_BLACKLISTITEM", "items")
end

Window_ItemTab_RemoveUnusedItems = function(widget, event)
	local self = AsmoRaidTracker
	
	self:DbRemoveUnusedItems()
	
	Window_OnSelectTab(self.displayWidgets.tabGroup, "ASMORAIDTRACKER_REMOVEUNUSEDITEMS", "items")
end

--------------------------------------------------------------------------------
-- Drawing items tab

function AsmoRaidTracker:DrawItemList(itemList)
	local items = self.db.profile.items
	local filterItemName = self.displayState.itemTab.filterItemName
	
	-- Create item list
	for _, item in ItemFilter(items, nil, filterItemName) do
		--------------------------
		-- the overlying group 
		local sgRow = GUI:Create("SimpleGroup")
		sgRow:SetLayout("AsmoRaidTracker_ItemRow")
		sgRow:SetFullWidth(true)
		sgRow:SetFullHeight(false)
		itemList:AddChild(sgRow)
		-- Item name
		local ilItemName = GUI:Create("InteractiveLabel")
		ilItemName:SetWidth(300)
		ilItemName:SetText(fmt("   |T%s:40:40:0:10|t  |cFF%s%s", item.itemTexture, RARITY_COLORS[item.itemRarity], item.itemName))
		local _, itemlink = GetItemInfo(item.itemId)
		if itemlink then
			ilItemName:SetUserData("itemlink", itemlink)
			ilItemName:SetCallback("OnEnter", Window_LootTab_ItemNameEnter)
			ilItemName:SetCallback("OnLeave", Window_LootTab_ItemNameLeave)
		end
		sgRow:AddChild(ilItemName)
		-- button remove item
		local buRemoveItem = GUI:Create("Button")
		buRemoveItem:SetWidth(100)
		buRemoveItem:SetText("Supprimer")
		buRemoveItem:SetCallback("OnClick", Window_ItemTab_RemoveItem)
		buRemoveItem:SetUserData("itemid", item.itemId)
		sgRow:AddChild(buRemoveItem)
		-- button blacklist item
		local buBlacklistItem = GUI:Create("Button")
		buBlacklistItem:SetWidth(100)
		buBlacklistItem:SetText("Liste noire")
		buBlacklistItem:SetCallback("OnClick", Window_ItemTab_BlacklistItem)
		buBlacklistItem:SetUserData("itemid", item.itemId)
		sgRow:AddChild(buBlacklistItem)
		
	end
end

function AsmoRaidTracker:DrawItemsTab(tabGroup)
    local items = self.db.profile.items
    
    tabGroup:SetLayout("Flow")
    
    ---------------------------------------
	-- Filter box
	local igFilter = GUI:Create("InlineGroup")
	igFilter:SetTitle(fmt("Items (%.0f)", items.length))
	igFilter:SetFullWidth(true)
	igFilter:SetLayout("Flow")
	igFilter:SetHeight(10)
	tabGroup:AddChild(igFilter)
	
	-- Item name filter
	local ebFilterByItemName = GUI:Create("EditBox")
	ebFilterByItemName:SetWidth(150)
	ebFilterByItemName:SetLabel("Filtre nom")
	ebFilterByItemName:SetText("")
	self.displayState.itemTab.filterItemName = ""
	ebFilterByItemName:SetCallback("OnTextChanged", Window_ItemTab_FilterItemName)
	igFilter:AddChild(ebFilterByItemName)
	
	-- button remove unused items
	local buRemoveUnusedItems = GUI:Create("Button")
	buRemoveUnusedItems:SetWidth(200)
	buRemoveUnusedItems:SetText("Supprimer items inutilisés")
	buRemoveUnusedItems:SetCallback("OnClick", Window_ItemTab_RemoveUnusedItems)
	igFilter:AddChild(buRemoveUnusedItems)
	
	---------------------------------------
	-- Loot list
	
	-- Scroll frame
	local scItemList = GUI:Create("ScrollFrame")
	self.displayWidgets.itemTab.itemList = scItemList
	scItemList:SetLayout("Flow")
	scItemList:SetFullWidth(true)
	scItemList:SetFullHeight(true)	
	tabGroup:AddChild(scItemList)
	
	self:DrawItemList(scItemList)
end

--------------------------------------------------------------------------------
-- Callbacks for export tab

--------------------------------------------------------------------------------
-- Drawing export tab

function AsmoRaidTracker:DrawExportTab(tabGroup, raidKey)
    local raid = self.db.profile.raids[raidKey]
    
    tabGroup:SetLayout("Flow")
    
    ---------------------------------------
	-- Raid members
	local mlMembers = GUI:Create("MultiLineEditBox")
	mlMembers:SetLabel("Membres")
	mlMembers:SetWidth(250)
	mlMembers:SetNumLines(15)
	local memberList = {}
	for _, member in pairs(raid.members) do
		tinsert(memberList, member.name)
	end
	mlMembers:SetText(tconcat(memberList, "\n"))
	tabGroup:AddChild(mlMembers)
	
	---------------------------------------
	-- Loots
	local mlLoots = GUI:Create("MultiLineEditBox")
	mlLoots:SetLabel("Loots")
	mlLoots:SetWidth(350)
	mlLoots:SetNumLines(15)
	local lootList = {}
	for _, loot in pairs(raid.loots) do
		if not (loot.looterName == "*DISENCHANT" or loot.looterName == "*LQE" or loot.looterName == "*FREE") then
			local item = self.db.profile.items[loot.itemId]
			if item then
				tinsert(lootList, fmt("%s:%s:%s", 
					loot.dkpValue or 0, 
					item.itemName, 
					loot.looterName
				))
			else
				tinsert(lootList, fmt("[Item id inconnu : %s]:%s", loot.itemId, loot.looterName))
			end
		end
	end
	mlLoots:SetText(tconcat(lootList, "\n"))
	tabGroup:AddChild(mlLoots)
	
end

--------------------------------------------------------------------------------
-- Callbacks for main window

-- Callbacks for Beginning/Ending record of a raid
Window_BeginRaid = function()
	local self = AsmoRaidTracker

	-- Display the popup to accept raid
	StaticPopup_Show("ASMORAIDTRACKER_CONFIRMRAID")
end
Window_RestartRecordingRaid = function()
	local self = AsmoRaidTracker
	local raidKey = self.displayState.selectedRaid

    AsmoRaidTracker:RestartRecordingRaid(raidKey)
	
	Window_OnClose(self.displayWidgets.window, "ASMORAIDTRACKER_RESTARTRECORDINGRAID")
	self:DrawWindow()
end
Window_EndRaid = function()
	local self = AsmoRaidTracker
	local raidKey = self.currentRaid.beginningTimestamp

	AsmoRaidTracker:EndRaid()
	
	Window_OnClose(self.displayWidgets.window, "ASMORAIDTRACKER_ENDRAID")
	self:DrawWindow()
end
Window_AssumeControlRaid = function()
	local self = AsmoRaidTracker
	local raid = self.currentRaid

    self.synchronizer:BroadcastRaidBegin(UnitName("player"), raid, self.currentRaid_MembersStatus)
	
	Window_OnClose(self.displayWidgets.window, "ASMORAIDTRACKER_ASSUMERAIDCONTROL")
	self:DrawWindow()
end

-- Callbacks for Removing raid
Window_DeleteRaid = function()
	local self = AsmoRaidTracker

	self:DbDeleteRaid(self.displayState.selectedRaid)
	
	Window_OnClose(self.displayWidgets.window, "ASMORAIDTRACKER_DELETERAID")
	self:DrawWindow()
end

Window_SendRaid = function()
	local self = AsmoRaidTracker
	local raid = self.db.profile.raids[self.displayState.selectedRaid]
	
	local items = {}
	for _, loot in pairs(raid.loots) do
		items[loot.itemId] = self.db.profile.items[loot.itemId]
	end
	
	self.synchronizer:BroadcastRaid(raid, items)
end

Window_RaidFilter = function(widget, event, raidKey)
	local self = AsmoRaidTracker
	self.displayState.selectedRaid = raidKey
	
	Window_OnClose(self.displayWidgets.window, "ASMORAIDTRACKER_DELETERAID")
	self:DrawWindow(raidKey)
end

Window_OnSelectTab = function(tabGroup, event, tabKey)
	local self = AsmoRaidTracker
	local selectedRaid = self.displayState.selectedRaid
	
	self.displayState.activeTab = tabKey
	
	tabGroup:ReleaseChildren()
	if tabKey == "raid" then
		self:DrawRaidTab(tabGroup, selectedRaid)
	elseif tabKey == "loots" then
	    self:DrawLootsTab(tabGroup, selectedRaid)
	elseif tabKey == "items" then
	    self:DrawItemsTab(tabGroup, selectedRaid)
	elseif tabKey == "export" then
	    self:DrawExportTab(tabGroup, selectedRaid)
	end
end

Window_OnClose = function(widget, event)
	local self = AsmoRaidTracker
	
	-- save display state
 	self.displayState.isWindowOpened = false
	
	AsmoRaidTracker_Window = nil
	for k, v in pairs(UISpecialFrames) do
		if v == "AsmoRaidTracker_Window" then
			tremove(UISpecialFrames, k)
		end
	end
	
	-- Save position, height and width
	local point, _, relativePoint, xOfs, yOfs = widget:GetPoint(1)
	local width = widget.frame:GetWidth()
	local height = widget.frame:GetHeight()
	self.db.profile.windowInfo = {}
	self.db.profile.windowInfo.point = point
	self.db.profile.windowInfo.relativePoint = relativePoint
	self.db.profile.windowInfo.xOfs = xOfs
	self.db.profile.windowInfo.yOfs = yOfs
	self.db.profile.windowInfo.width = width
	self.db.profile.windowInfo.height = height
	
	-- Release frame
	GUI:Release(widget)
end

--------------------------------------------------------------------------------
-- Drawing main windows

function AsmoRaidTracker:DrawTabGroup(raidKey)
	if raidKey then
		-- Create a tab group
		local tg = GUI:Create("TabGroup")
		self.displayWidgets.tabGroup = tg
		tg:SetFullWidth(true)
		tg:SetFullHeight(true)
		tg:SetTabs(RAIDWINDOW_TABLIST)
		tg:SetCallback("OnGroupSelected", Window_OnSelectTab)
		tg:SelectTab(self.displayState.activeTab or "raid")
	 	self.displayWidgets.window:AddChild(tg)
	end
end

function AsmoRaidTracker:DrawWindow(raidKey)
	-- save display state
 	self.displayState.isWindowOpened = true
 	
	-- Create container frame
	local f = GUI:Create("Frame")
	self.displayWidgets.window = f
	f:SetCallback("OnClose", Window_OnClose)
	f:SetTitle("Asmo Raid Tracker")
	f:SetStatusText("")
	f:SetLayout("Flow")
	do
		local info = self.db.profile.windowInfo
		if info then
			f:SetPoint(info.point, UIParent, info.relativePoint, info.xOfs, info.yOfs)
			f:SetWidth(info.width)
			f:SetHeight(info.height)
		end
	end
	f:Show()
	
	-- Add to escape-capable frames 
	AsmoRaidTracker_Window = f
	tinsert(UISpecialFrames, "AsmoRaidTracker_Window")
	
	-- Create raid dropdown list
	local raidList, firstRaidKey = self:DbGetRaidListForDropdown()
	if not raidKey then raidKey = firstRaidKey end
	local ddRaid = GUI:Create("Dropdown")
	ddRaid:SetWidth(250)
	ddRaid:SetList(raidList)
	ddRaid:SetLabel("Raid")
	ddRaid:SetCallback("OnValueChanged", Window_RaidFilter)
	ddRaid:SetValue(raidKey)
	self.displayState.selectedRaid = raidKey
	f:AddChild(ddRaid)
	
	-- Control buttons
	-- If there is a raid selected
	if raidKey then
		local raid = self.db.profile.raids[raidKey]
		-- If selected raid is the ongoing raid and user is the master and raid is not finished, user can finish it
		if self.currentRaid == raid and not self.synchronizer.masterName and not raid.endTimestamp then
			-- End raid
			local buEndRaid = GUI:Create("Button")
			buEndRaid:SetWidth(150)
			buEndRaid:SetText("Finir")
			buEndRaid:SetCallback("OnClick", Window_EndRaid)
			f:AddChild(buEndRaid)
		end
		-- If selected raid is not finished and there is no ongoing raid, user can restart recording on it
		if not raid.endTimestamp and not self.currentRaid then
			-- Reprendre raid
			local buRestartRecordingRaid = GUI:Create("Button")
			buRestartRecordingRaid:SetWidth(150)
			buRestartRecordingRaid:SetText("Reprendre")
			buRestartRecordingRaid:SetCallback("OnClick", Window_RestartRecordingRaid)
			f:AddChild(buRestartRecordingRaid)
		end
		-- If there is no ongoing raid, user can begin one
		if not self.currentRaid then
			-- Begin raid
			local buBeginRaid = GUI:Create("Button")
			buBeginRaid:SetWidth(150)
			buBeginRaid:SetText("Commencer")
			buBeginRaid:SetCallback("OnClick", Window_BeginRaid)
			f:AddChild(buBeginRaid)
		end
		-- If selected raid is not ongoing, user can delete it
		if self.currentRaid ~= raid then
			-- Delete raid
			local buDeleteRaid = GUI:Create("Button")
			buDeleteRaid:SetWidth(150)
			buDeleteRaid:SetText("Supprimer")
			buDeleteRaid:SetCallback("OnClick", Window_DeleteRaid)
			f:AddChild(buDeleteRaid)
			-- Send raid
			local buSendRaid = GUI:Create("Button")
			buSendRaid:SetWidth(150)
			buSendRaid:SetText("Synchroniser")
			buSendRaid:SetCallback("OnClick", Window_SendRaid)
			f:AddChild(buSendRaid)
		end
		-- If selected raid is the ongoing raid and user is not the master, user can assume control of the raid
		if self.currentRaid == raid and self.synchronizer.masterName then
			-- Assume control of raid
			local buAssumeControl = GUI:Create("Button")
			buAssumeControl:SetWidth(150)
			buAssumeControl:SetText("Prendre contrôle")
			buAssumeControl:SetCallback("OnClick", Window_AssumeControlRaid)
			f:AddChild(buAssumeControl)
		end 
	else
		-- Begin raid
		local buBeginRaid = GUI:Create("Button")
		buBeginRaid:SetWidth(150)
		buBeginRaid:SetText("Commencer")
		buBeginRaid:SetCallback("OnClick", Window_BeginRaid)
		f:AddChild(buBeginRaid)
	end
	
	
	-- Create tab group
	self:DrawTabGroup(raidKey)
end

--------------------------------------------------------------------------------
-- Update display methods

-- Called when an event external event might update current raid panel
function AsmoRaidTracker:UpdateDisplay_SelectedRaid(event)
	-- If window is opened
	if self.displayState.isWindowOpened then
		Window_OnClose(self.displayWidgets.window, event)
		self:DrawWindow()
	end
end

function AsmoRaidTracker:UpdateDisplay_RaidRoster(raidKey)
	-- Update raid tab if opened and displayed raid is current raid
	if self.displayState.isWindowOpened and self.displayState.activeTab == "raid" and raidKey == self.displayState.selectedRaid then
        local tabGroup = self.displayWidgets.tabGroup
        tabGroup:ReleaseChildren()
		self:DrawRaidTab(tabGroup, raidKey)
	end
end

function AsmoRaidTracker:UpdateDisplay_LootList(raidKey)
    -- If window opened
    if self.displayState.isWindowOpened then
    	-- If loot tab is opened on current raid
    	if self.displayState.activeTab == "loots" and raidKey == self.displayState.selectedRaid then
			local lootList = self.displayWidgets.lootTab.lootList
	        lootList:ReleaseChildren()
	  		self:DrawLootList(lootList, raidKey)
	  	-- Else if item tab is opened
	  	elseif self.displayState.activeTab == "items" then
	  		Window_OnSelectTab(self.displayWidgets.tabGroup, "ASMORAIDTRACKER_UPDATELOOTS", "items")
		end
	end
end