local AI = AsmoRaidTracker:NewModule("Inviter", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceBucket-3.0")
local RAIDROSTER_DELAY = 2

--------------------------------------------------------------------------------
-- Helpers

-- Aliases
local tconcat, tinsert = table.concat, table.insert
local fmt = string.format

-- allows descending sort
local function descSorter(a, b)
	return a >= b
end

-- allows sort of raider by class
local function classSorter(a, b)
	aClass = AI.classInfo[a]
	bClass = AI.classInfo[b]
	if aClass and bClass then
		if aClass == bClass then
			return a < b
		else
			return aClass < bClass
		end
	else
		return a < b
	end
end

-- allows to iterate a table based on key alphabetical order
local function pairsByKey (t, f)
	local a = {}
	for n in pairs(t) do table.insert(a, n) end
	table.sort(a, f)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], t[a[i]]
		end
	end
	return iter
end

-- Raid iterator
local function RaidIterator()
	local i = 0
	return function ()
		i = i + 1
		local unit
		if UnitInRaid("player") and i<=GetNumRaidMembers() then
			return "raid"..i
		else
			numPartyMembers = GetNumPartyMembers()
			if i<= numPartyMembers then
				return "party"..i
			elseif i==numPartyMembers+1 then
				return "player"
			end
		end
	end
end

-- allows to iterate raiders
local function RaiderFilter (raiders, sorter, excludeOffline, excludeGrouped, includeTank,	includeHeal, includeMelee, includeRanged)
	local a = {}
	for raiderName, raider in pairs(raiders) do
		local excluded = false
		-- On récupère les gens online dans la guilde
		local onlineMembers = AI:GetOnlineGuildMembers()
		-- Si le gars est offline, on l'exclut d'office
		if excludeOffline and not onlineMembers[raiderName] then
			excluded = true
		end
		-- On récupère les gens groupés 
		local groupedMembers = AI:GetAlreadyGroupedMembers()
		-- Si le gars est déjà groupé on l'exclut
		if excludeGrouped and groupedMembers[raiderName] then
			excluded = true
		end
		-- Pour les gens non exclus
		if not excluded then
			local included = false
			if includeTank and raider.isTank then
				included = true
			end
			if includeHeal and raider.isHeal then
				included = true
			end
			if includeMelee and raider.isMelee then
				included = true
			end
			if includeRanged and raider.isRanged then
				included = true
			end
			-- Les gars qui n'ont aucun rôle sont inclus par défaut
			if not raider.isTank and not raider.isHeal and not raider.isMelee and not raider.isRanged then
				included = true
			end
			if included then
				tinsert(a, raiderName)
			end
		end 
	end
	table.sort(a, sorter)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], raiders[a[i]]
		end
	end
	return iter
end

local function getBeginningDayTimestamp()
	local timestamp = time()
	local h = date("%H", timestamp)
	local m = date("%M", timestamp)
	local s = date("%S", timestamp)
	return timestamp-h*3600-m*60-s+5*3600
end

--------------------------------------------------------------------------------
-- Libraries instanciation
local GUI = LibStub("AceGUI-3.0")

--------------------------------------------------------------------------------
-- Callbacks declaration
local WindowInviter_OnClose
local WindowInviter_AddRaiderIncludeOffline
local WindowInviter_AddRaiderSelectGuildMember
local WindowInviter_AddRaider
local WindowInviter_ApplyFilter
local WindowInviter_SynchronizeRoster
local WindowInviter_ToggleRaiderRole

--------------------------------------------------------------------------------
-- Standard Addon methods

function AI:OnInitialize()

end

function AI:OnEnable()
	-- Setting up
    AsmoRaidTracker.inviter = self
    self.displayState = {
    	filtersIncludeTank = true,
    	filtersIncludeHeal = true,
    	filtersIncludeMelee = true,
    	filtersIncludeRanged = true,
	}
    self.displayWidgets = {}
    self.classInfo = {}
	
    -- Register roster changes 
    self.raidRosterUpdateBucketEvent = self:RegisterBucketEvent("PARTY_MEMBERS_CHANGED", RAIDROSTER_DELAY, "OnRosterUpdate")
    -- Register log on event (guild roster update to ensure that guild roster info is accurate) 
    self:RegisterEvent("GUILD_ROSTER_UPDATE", "OnGuildRosterReady")
    
    self:ComputeStats()
end

function AI:OnDisable()
    -- Called when the addon is disabled
	
end

--------------------------------------------------------------------------------
-- Methods

function AI:OnGuildRosterReady()
	local index = 1	
	local name, _, _, _, _, _, _, _, _, _, class = GetGuildRosterInfo(index) 
	while name do
		self.classInfo[name] = class
		index = index + 1
		name, _, _, _, _, _, _, _, _, _, class = GetGuildRosterInfo(index)		
	end
end

function AI:GetEligibleGuildMembers(offlineIncluded)
	GuildRoster()
	local raiders = AsmoRaidTracker.db.profile.raiders
	local excludedGuildRanks = AsmoRaidTracker.db.profile.excludedGuildRanks
	-- construct list
	local eligibles = {}
	local index = 1	
	local name, _, rankIndex, _, _, _, _, _, online = GetGuildRosterInfo(index) 
	while name do
		-- If offline are included, we consider everyone as online
		if offlineIncluded then online = true end
		-- If member is online, not in raid and not part of an excluded rank		
		if online and not raiders[name] and not excludedGuildRanks[rankIndex] then 		
			eligibles[name] = name
		end
		index = index + 1
		name, rank, rankIndex, level, _, _, _, _, online = GetGuildRosterInfo(index)		
	end
	
	return eligibles
end

function AI:GetOnlineGuildMembers()
	GuildRoster()
	local onlineList = {}
	local index = 1	
	local name, _, rankIndex, _, _, _, _, _, online = GetGuildRosterInfo(index) 
	while name do
		-- If member is online, not in raid and not part of an excluded rank		
		if online then 		
			onlineList[name] = true
		end
		index = index + 1
		name, rank, rankIndex, level, _, _, _, _, online = GetGuildRosterInfo(index)		
	end
	
	return onlineList
end

function AI:GetAlreadyGroupedMembers()
	local raiders = {}
	for unit in RaidIterator() do
		raiders[UnitName(unit)] = true
	end
	return raiders
end

function AI:ComputeStats()
	local currentTimestamp = getBeginningDayTimestamp()
	local stats = {}
	local totalRaidNumber = 0
	
	local lastRaid
	for _, raid in pairsByKey(AsmoRaidTracker.db.profile.raids, descSorter) do
		if not lastRaid then lastRaid = raid end
		-- We only take into account data less than 31 days
		if raid.beginningTimestamp < currentTimestamp - 31*24*3600 then break end
		if raid ~= AsmoRaidTracker.currentRaid then -- Skip current raid if any
			totalRaidNumber = totalRaidNumber + 1
			
			for _, member in pairs(raid.members) do
				local stat = stats[member.name]
				if not stat then
					stat = {
						numberRaid = 0,
						numberPrio30 = 0,
						numberPrio7 = 0,
						numberPrio1 = 0,
						numberPrioLast = 0,
					}
					stats[member.name] = stat
				end
				stat.numberRaid = stat.numberRaid + 1
				if member.isPrio then
					stat.numberPrio30 = stat.numberPrio30 + 1
					if raid.beginningTimestamp >= currentTimestamp - 7*24*3600 then
						stat.numberPrio7 = stat.numberPrio7 + 1
					end
					if date("%w", raid.beginningTimestamp) == date("%w", currentTimestamp) then
						stat.numberPrio1 = stat.numberPrio1 + 1
					end
				end
			end
		end
	end
	
	for _, member in pairs(lastRaid.members) do
		local stat = stats[member.name]
		if member.isPrio then
			stat.numberPrioLast = 1
		end
	end
	
	
	for _, stat in pairs(stats) do
		if totalRaidNumber > 0 then
			stat.attendance = stat.numberRaid / totalRaidNumber
		end
	end
	
	self.stats = stats	
end

--------------------------------------------------------------------------------
-- Events handling

function AI:OnRosterUpdate()
	if self.displayState.isWindowOpened	then
		WindowInviter_OnClose(self.displayWidgets.window)
		self:DrawWindow()
	end 
end

--------------------------------------------------------------------------------
-- Callbacks for main window

WindowInviter_OnClose = function(widget, event)
	local self = AI
	
	-- save display state
 	self.displayState.isWindowOpened = false
	
	AsmoRaidTracker_InviterWindow = nil
	for k, v in pairs(UISpecialFrames) do
		if v == "AsmoRaidTracker_InviterWindow" then
			tremove(UISpecialFrames, k)
		end
	end
	
	-- Save position, height and width
	local point, _, relativePoint, xOfs, yOfs = widget:GetPoint(1)
	local width = widget.frame:GetWidth()
	local height = widget.frame:GetHeight()
	local info = {}
	info.point = point
	info.relativePoint = relativePoint
	info.xOfs = xOfs
	info.yOfs = yOfs
	info.width = width
	info.height = height
	AsmoRaidTracker.db.profile.windowInviterInfo = info
	
	-- Release frame
	GUI:Release(widget)
end

-- Add raiders callback
WindowInviter_AddRaiderIncludeOffline = function(widget, event, v)
	local self = AI
	
	self.displayState.addRaiderIncludeOffline = v
	self.displayWidgets.addRaiderSelectMember:SetList(self:GetEligibleGuildMembers(v))
end
WindowInviter_AddRaiderSelectGuildMember = function(widget, event, v)
    local self = AI
	    
    self.displayState.addRaiderGuildMember = v
end
WindowInviter_AddRaider = function(widget, event)
	local self = AI
	local raiderName = self.displayState.addRaiderGuildMember
	
	if raiderName then
		AsmoRaidTracker:DbAddRaiderToRoster(raiderName)
		self.displayState.addRaiderGuildMember = nil
	end
	-- Update display
	WindowInviter_OnClose(self.displayWidgets.window)
	self:DrawWindow()
end

-- Filters callback
WindowInviter_ApplyFilter = function(widget, event)
    local self = AI
    
    self.displayWidgets.raiderList:ReleaseChildren()
    self:DrawRaiderList()
end

-- Synchronisation du roster
WindowInviter_SynchronizeRoster = function(widget, event)
	AsmoRaidTracker.synchronizer:BroadcastRaidRoster()
end

-- Toggle raider role 
WindowInviter_ToggleRaiderRole = function(widget, event, v)
	local self = AI
	local raiders = AsmoRaidTracker.db.profile.raiders
	local raiderName = widget:GetUserData("raiderName")
	local role = widget:GetUserData("role")
	
	raiders[raiderName][role] = v
	
	-- Update display
	WindowInviter_OnClose(self.displayWidgets.window)
	self:DrawWindow()
end

--------------------------------------------------------------------------------
-- Drawing main windows

-- Draw main windown
function AI:DrawWindow()
	-- save display state
 	self.displayState.isWindowOpened = true
 	
	-- Create container frame
	local f = GUI:Create("Frame")
	self.displayWidgets.window = f
	f:SetCallback("OnClose", WindowInviter_OnClose)
	f:SetTitle("ART - Gestion invites")
	f:SetStatusText("")
	f:SetLayout("Flow")
	do
		local info = AsmoRaidTracker.db.profile.windowInviterInfo
		if info then
			f:SetPoint(info.point, UIParent, info.relativePoint, info.xOfs, info.yOfs)
			f:SetWidth(info.width)
			f:SetHeight(info.height)
		end
	end
	f:Show()
	
	-- Add to escape-capable frames 
	AsmoRaidTracker_InviterWindow = f
	tinsert(UISpecialFrames, "AsmoRaidTracker_InviterWindow")
	
	----------------------------------------------------------------------------
	-- Current raid composition
	local gRaidComp = GUI:Create("SimpleGroup")
	gRaidComp:SetFullWidth(true)
	gRaidComp:SetLayout("Flow")
	f:AddChild(gRaidComp)
	
	local raiders = AsmoRaidTracker.db.profile.raiders
	local tankList, healList, meleeList, rangedList = {}, {}, {}, {}
	for unit in RaidIterator() do
		local raider = raiders[UnitName(unit)]
		if raider then
			if raider.isTank then
				if raider.isHeal or raider.isMelee or raider.isRanged then
					tinsert(tankList, "*"..raider.name)
				else
					tinsert(tankList, raider.name)
				end
			end
			if raider.isHeal then
				if raider.isTank or raider.isMelee or raider.isRanged then
					tinsert(healList, "*"..raider.name)
				else
					tinsert(healList, raider.name)
				end
			end
			if raider.isMelee then
				if raider.isHeal or raider.isTank or raider.isRanged then
					tinsert(meleeList, "*"..raider.name)
				else
					tinsert(meleeList, raider.name)
				end
			end
			if raider.isRanged then
				if raider.isHeal or raider.isTank or raider.isTank then
					tinsert(rangedList, "*"..raider.name)
				else
					tinsert(rangedList, raider.name)
				end
			end
		else
			f:SetStatusText(fmt("/!\\ Warning : %s n'est pas dans la liste des raiders", UnitName(unit)))
		end
	end	
	-- Tank list
	local mlTanks = GUI:Create("MultiLineEditBox")
	mlTanks:SetLabel("Tanks")
	mlTanks:SetWidth(160)
	mlTanks:SetNumLines(10)
	mlTanks:SetDisabled(true)
	mlTanks:SetText(tconcat(tankList, "\n"))
	gRaidComp:AddChild(mlTanks)	
	-- Heal list
	local mlHeals = GUI:Create("MultiLineEditBox")
	mlHeals:SetLabel("Heals")
	mlHeals:SetWidth(160)
	mlHeals:SetNumLines(10)
	mlHeals:SetDisabled(true)
	mlHeals:SetText(tconcat(healList, "\n"))
	gRaidComp:AddChild(mlHeals)
	-- Melee list
	local mlMelees = GUI:Create("MultiLineEditBox")
	mlMelees:SetLabel("CaC")
	mlMelees:SetWidth(160)
	mlMelees:SetNumLines(10)
	mlMelees:SetDisabled(true)
	mlMelees:SetText(tconcat(meleeList, "\n"))
	gRaidComp:AddChild(mlMelees)
	-- Ranged list
	local mlRanged = GUI:Create("MultiLineEditBox")
	mlRanged:SetLabel("Ranged")
	mlRanged:SetWidth(160)
	mlRanged:SetNumLines(10)
	mlRanged:SetDisabled(true)
	mlRanged:SetText(tconcat(rangedList, "\n"))
	gRaidComp:AddChild(mlRanged)

	-------------------------------
	-- Add members	
	local igAddRaider = GUI:Create("InlineGroup")
	igAddRaider:SetTitle("Ajouter raider")
	igAddRaider:SetWidth(300)
	igAddRaider:SetLayout("Flow")
	f:AddChild(igAddRaider)
	
	-- Include offline checkbox
	local cbAddRaiderIncludeOffline = GUI:Create("CheckBox")
	cbAddRaiderIncludeOffline:SetWidth(75)
	cbAddRaiderIncludeOffline:SetType("checkbox")
	cbAddRaiderIncludeOffline:SetLabel("Offline")
	cbAddRaiderIncludeOffline:SetValue(false)
	self.displayState.addRaiderIncludeOffline = false
	cbAddRaiderIncludeOffline:SetTriState(false)
	cbAddRaiderIncludeOffline:SetCallback("OnValueChanged", WindowInviter_AddRaiderIncludeOffline)
	igAddRaider:AddChild(cbAddRaiderIncludeOffline)
	-- Guild member select
	local ddAddRaider = GUI:Create("Dropdown")
	self.displayWidgets.addRaiderSelectMember = ddAddRaider
	ddAddRaider:SetWidth(150)
	ddAddRaider:SetList(self:GetEligibleGuildMembers(false))
	ddAddRaider:SetLabel("Membres")
	ddAddRaider:SetText("Sélectionner")
	self.displayState.addRaiderGuildMember = nil
	ddAddRaider:SetCallback("OnValueChanged", WindowInviter_AddRaiderSelectGuildMember)
	igAddRaider:AddChild(ddAddRaider)
	-- OK button
	local buAddRaiderConfirm = GUI:Create("Button")
	buAddRaiderConfirm:SetWidth(50)
	buAddRaiderConfirm:SetText("OK")
	buAddRaiderConfirm:SetCallback("OnClick", WindowInviter_AddRaider)
	igAddRaider:AddChild(buAddRaiderConfirm)
	
	-------------------------------
	-- Filtres	
	local igFilters = GUI:Create("InlineGroup")
	igFilters:SetTitle("Filtres")
	igFilters:SetWidth(340)
	igFilters:SetLayout("Flow")
	f:AddChild(igFilters)
	
	-- Filter offline
	local cbFilterIncludeOffline = GUI:Create("CheckBox")
	cbFilterIncludeOffline:SetWidth(100)
	cbFilterIncludeOffline:SetType("checkbox")
	cbFilterIncludeOffline:SetLabel("Offline")
	cbFilterIncludeOffline:SetValue(self.displayState.filtersExcludeOffline or false)
	cbFilterIncludeOffline:SetTriState(false)
	cbFilterIncludeOffline:SetCallback("OnValueChanged", function(widget, event, v)
		local self = AI
		
		self.displayState.filtersExcludeOffline = v
		WindowInviter_ApplyFilter()
	end)
	igFilters:AddChild(cbFilterIncludeOffline)
	-- Filter already in raid
	local cbFilterIncludeUngrouped = GUI:Create("CheckBox")
	cbFilterIncludeUngrouped:SetWidth(100)
	cbFilterIncludeUngrouped:SetType("checkbox")
	cbFilterIncludeUngrouped:SetLabel("Groupés")
	cbFilterIncludeUngrouped:SetValue(self.displayState.filtersExcludeGrouped or false)
	cbFilterIncludeUngrouped:SetTriState(false)
	cbFilterIncludeUngrouped:SetCallback("OnValueChanged", function(widget, event, v)
		local self = AI
		
		self.displayState.filtersExcludeGrouped = v
		WindowInviter_ApplyFilter()
	end)
	igFilters:AddChild(cbFilterIncludeUngrouped)
	-- Filter tanks
	local cbFilterIncludeTank = GUI:Create("CheckBox")
	cbFilterIncludeTank:SetWidth(100)
	cbFilterIncludeTank:SetType("checkbox")
	cbFilterIncludeTank:SetLabel("Tanks")
	cbFilterIncludeTank:SetValue(self.displayState.filtersIncludeTank or false)
	cbFilterIncludeTank:SetTriState(false)
	cbFilterIncludeTank:SetCallback("OnValueChanged", function(widget, event, v)
		local self = AI
		
		self.displayState.filtersIncludeTank = v
		WindowInviter_ApplyFilter()
	end)
	igFilters:AddChild(cbFilterIncludeTank)
	-- Filter heals
	local cbFilterIncludeHeal = GUI:Create("CheckBox")
	cbFilterIncludeHeal:SetWidth(100)
	cbFilterIncludeHeal:SetType("checkbox")
	cbFilterIncludeHeal:SetLabel("Heals")
	cbFilterIncludeHeal:SetValue(self.displayState.filtersIncludeHeal or false)
	cbFilterIncludeHeal:SetTriState(false)
	cbFilterIncludeHeal:SetCallback("OnValueChanged", function(widget, event, v)
		local self = AI
		
		self.displayState.filtersIncludeHeal = v
		WindowInviter_ApplyFilter()
	end)
	igFilters:AddChild(cbFilterIncludeHeal)
	-- Filter melee
	local cbFilterIncludeMelee = GUI:Create("CheckBox")
	cbFilterIncludeMelee:SetWidth(100)
	cbFilterIncludeMelee:SetType("checkbox")
	cbFilterIncludeMelee:SetLabel("CaC")
	cbFilterIncludeMelee:SetValue(self.displayState.filtersIncludeMelee or false)
	cbFilterIncludeMelee:SetTriState(false)
	cbFilterIncludeMelee:SetCallback("OnValueChanged", function(widget, event, v)
		local self = AI
		
		self.displayState.filtersIncludeMelee = v
		WindowInviter_ApplyFilter()
	end)
	igFilters:AddChild(cbFilterIncludeMelee)
	-- Filter ranged
	local cbFilterIncludeRanged = GUI:Create("CheckBox")
	cbFilterIncludeRanged:SetWidth(75)
	cbFilterIncludeRanged:SetType("checkbox")
	cbFilterIncludeRanged:SetLabel("Ranged")
	cbFilterIncludeRanged:SetValue(self.displayState.filtersIncludeRanged or false)
	cbFilterIncludeRanged:SetTriState(false)
	cbFilterIncludeRanged:SetCallback("OnValueChanged", function(widget, event, v)
		local self = AI
		
		self.displayState.filtersIncludeRanged = v
		WindowInviter_ApplyFilter()
	end)
	igFilters:AddChild(cbFilterIncludeRanged)

    -------------------------------
	-- General purpose actions
	local sgActions = GUI:Create("SimpleGroup")
	sgActions:SetFullWidth(true)
	sgActions:SetLayout("Flow")
	f:AddChild(sgActions)

	-- Refresh button
	local buRefresh = GUI:Create("Button")
	buRefresh:SetWidth(100)
	buRefresh:SetText("Rafraîchir")
	buRefresh:SetCallback("OnClick", WindowInviter_ApplyFilter)
	sgActions:AddChild(buRefresh)
	-- Synchronize button
	local buSynchronize = GUI:Create("Button")
	buSynchronize:SetWidth(100)
	buSynchronize:SetText("Synchro")
	buSynchronize:SetCallback("OnClick", WindowInviter_SynchronizeRoster)
	sgActions:AddChild(buSynchronize)
	
	----------------------------------------------------------------------------
	-- Raider list
	
	-- Column titles
	local sgColTitles = GUI:Create("SimpleGroup")
	sgColTitles:SetLayout("Flow")
	sgColTitles:SetFullWidth(true)
	f:AddChild(sgColTitles)
	-- Filler		
	local lFiller = GUI:Create("Label")
	lFiller:SetWidth(40)
	lFiller:SetText("")
	sgColTitles:AddChild(lFiller)
	-- Name		
	local lName = GUI:Create("Label")
	lName:SetWidth(100)
	lName:SetText("Nom")
	lName:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lName)
	-- Tank		
	local lTank = GUI:Create("Label")
	lTank:SetWidth(20)
	lTank:SetText("   T")
	lTank:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lTank)
	-- Heal		
	local lHeal = GUI:Create("Label")
	lHeal:SetWidth(20)
	lHeal:SetText("   H")
	lHeal:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lHeal)
	-- Melee		
	local lHeal = GUI:Create("Label")
	lHeal:SetWidth(20)
	lHeal:SetText("   C")
	lHeal:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lHeal)
	-- Ranged		
	local lHeal = GUI:Create("Label")
	lHeal:SetWidth(30)
	lHeal:SetText("   R")
	lHeal:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lHeal)
	-- Attendance		
	local lAttendance = GUI:Create("Label")
	lAttendance:SetWidth(60)
	lAttendance:SetText("Attendance (30j)")
	lAttendance:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lAttendance)
	-- Prios (30j)		
	local lPrio30 = GUI:Create("Label")
	lPrio30:SetWidth(60)
	lPrio30:SetText("Prios (30j)")
	lPrio30:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lPrio30)
	-- Prios (7j)		
	local lPrio7 = GUI:Create("Label")
	lPrio7:SetWidth(60)
	lPrio7:SetText("Prios (7j)")
	lPrio7:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lPrio7)
	-- Prios (même jour)		
	local lPrio1 = GUI:Create("Label")
	lPrio1:SetWidth(60)
	lPrio1:SetText(fmt("Prios %s (30j)", date("%a", time())))
	lPrio1:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lPrio1)
	-- Prios (last raid)		
	local lPrio1 = GUI:Create("Label")
	lPrio1:SetWidth(60)
	lPrio1:SetText("Prio last")
	lPrio1:SetColor(255/255,230/255,0)
	sgColTitles:AddChild(lPrio1)
	
	-- Scroll frame
	local scRaiderList = GUI:Create("ScrollFrame")
	self.displayWidgets.raiderList = scRaiderList
	scRaiderList:SetLayout("Flow")
	scRaiderList:SetFullWidth(true)
	scRaiderList:SetFullHeight(true)	
	f:AddChild(scRaiderList)
	
	self:DrawRaiderList()
end

function AI:DrawRaiderList()
	local raiders = AsmoRaidTracker.db.profile.raiders
	local scRaiderList = self.displayWidgets.raiderList	
	
	for _, raider in RaiderFilter(raiders, classSorter,
		self.displayState.filtersExcludeOffline,
		self.displayState.filtersExcludeGrouped,
		self.displayState.filtersIncludeTank,
		self.displayState.filtersIncludeHeal,
		self.displayState.filtersIncludeMelee,
		self.displayState.filtersIncludeRanged 
	) do
		local stat = self.stats[raider.name]
	    --------------------------
		-- the overlying group 
		local sgRow = GUI:Create("SimpleGroup")
		sgRow:SetLayout("Flow")
		sgRow:SetFullWidth(true)
		scRaiderList:AddChild(sgRow)
		
		-- Delete button
		local buDeleteRaider = GUI:Create("Button")
		buDeleteRaider:SetWidth(40)
		buDeleteRaider:SetText("X")
		buDeleteRaider:SetUserData("raiderName", raider.name)
		buDeleteRaider:SetCallback("OnClick", function(widget, event)
			local self = AI
			local raiders = AsmoRaidTracker.db.profile.raiders
			local raiderName = widget:GetUserData("raiderName")
			
			raiders[raiderName] = nil
			
			-- Update display
			WindowInviter_OnClose(self.displayWidgets.window)
			self:DrawWindow()
		end)
		sgRow:AddChild(buDeleteRaider)
		-- Name		
		local lName = GUI:Create("Label")
		lName:SetWidth(100)
		lName:SetText(raider.name)
		if self.classInfo[raider.name] then
			local color = RAID_CLASS_COLORS[self.classInfo[raider.name]]
			lName:SetColor(color.r, color.g, color.b)
		end
		sgRow:AddChild(lName)
		-- Tank checkbox		
		local cbTank = GUI:Create("CheckBox")
		cbTank:SetType("checkbox")
		cbTank:SetWidth(20)
		cbTank:SetValue(raider.isTank)
		cbTank:SetUserData("raiderName", raider.name)
		cbTank:SetUserData("role", "isTank")
		cbTank:SetCallback("OnValueChanged", WindowInviter_ToggleRaiderRole)
		sgRow:AddChild(cbTank)
		-- Heal checkbox		
		local cbHeal = GUI:Create("CheckBox")
		cbHeal:SetType("checkbox")
		cbHeal:SetWidth(20)
		cbHeal:SetValue(raider.isHeal)
		cbHeal:SetUserData("raiderName", raider.name)
		cbHeal:SetUserData("role", "isHeal")
		cbHeal:SetCallback("OnValueChanged", WindowInviter_ToggleRaiderRole)
		sgRow:AddChild(cbHeal)
		-- Melee checkbox		
		local cbHeal = GUI:Create("CheckBox")
		cbHeal:SetType("checkbox")
		cbHeal:SetWidth(20)
		cbHeal:SetValue(raider.isMelee)
		cbHeal:SetUserData("raiderName", raider.name)
		cbHeal:SetUserData("role", "isMelee")
		cbHeal:SetCallback("OnValueChanged", WindowInviter_ToggleRaiderRole)
		sgRow:AddChild(cbHeal)
		-- Ranged checkbox		
		local cbHeal = GUI:Create("CheckBox")
		cbHeal:SetType("checkbox")
		cbHeal:SetWidth(30)
		cbHeal:SetValue(raider.isRanged)
		cbHeal:SetUserData("raiderName", raider.name)
		cbHeal:SetUserData("role", "isRanged")
		cbHeal:SetCallback("OnValueChanged", WindowInviter_ToggleRaiderRole)
		sgRow:AddChild(cbHeal)
		-- Attendance		
		local lAttendance = GUI:Create("Label")
		lAttendance:SetWidth(60)
		if stat then
			lAttendance:SetText(fmt("              %.0f%%", stat.attendance*100))
			if stat.attendance < 0.75 then
				lAttendance:SetColor(1, 0, 0)
			end
		end
		sgRow:AddChild(lAttendance)
		-- Prios (30j)		
		local lPrio30 = GUI:Create("Label")
		lPrio30:SetWidth(60)
		if stat then
			if stat.numberPrio30 > 0 then
				lPrio30:SetText(fmt("              %.0f", stat.numberPrio30))
			else
				lPrio30:SetText("")
			end
		end
		sgRow:AddChild(lPrio30)
		-- Prios (7j)		
		local lPrio7 = GUI:Create("Label")
		lPrio7:SetWidth(60)
		if stat then
			if stat.numberPrio7 > 0 then
				lPrio7:SetText(fmt("              %.0f", stat.numberPrio7))
			else
				lPrio7:SetText("")
			end
		end
		sgRow:AddChild(lPrio7)
		-- Prios (même jour)		
		local lPrio1 = GUI:Create("Label")
		lPrio1:SetWidth(60)
		if stat then
			if stat.numberPrio1 > 0 then
				lPrio1:SetText(fmt("              %.0f", stat.numberPrio1))
			else
				lPrio1:SetText("")
			end
		end
		sgRow:AddChild(lPrio1)
		-- Prios (last raid)		
		local lPrioLast = GUI:Create("Label")
		lPrioLast:SetWidth(60)
		if stat then
			if stat.numberPrioLast > 0 then
				lPrioLast:SetText(fmt("              %.0f", stat.numberPrioLast))
			else
				lPrioLast:SetText("")
			end
		end
		sgRow:AddChild(lPrioLast)
		-- Invite button
		local buInvite = GUI:Create("Button")
		buInvite:SetWidth(75)
		buInvite:SetText("Invite")
		buInvite:SetUserData("raiderName", raider.name)
		buInvite:SetCallback("OnClick", function(widget, event)
			local raiderName = widget:GetUserData("raiderName")
			InviteUnit(raiderName)
		end)
		sgRow:AddChild(buInvite) 
	end	
end