local AS = AsmoRaidTracker:NewModule("Synchronizer", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceSerializer-3.0")
local BROADCASTRAIDUPDATE_DELAY = 30
local BROADCASTRAIDUPDATE_TIMEOUT = 10

local libZip = LibStub:GetLibrary("LibCompress")

--------------------------------------------------------------------------------
-- Helpers

-- Aliases
local fmt = string.format
local tconcat, tinsert = table.concat, table.insert

-- Copy a var which can be a table or another simple type (not function) per value
local function copyvalue(t)
	if type(t) == "table" then
		local r = {}
		for k,v in pairs(t) do
			r[k]=copyvalue(v)
		end
		return r
	else
		return t
	end

end

-- takes a table as entry, return table where corresponding entries are transformed into strings to display
local function ttrunc(t)
	r = {}
	for k, v in pairs(t) do
	    local type = type(v)
	    if type == "table" then
	    	r[k] = "table"
		else
		    r[k] = v
		end
	end
	return r
end

-- allows descending sort
local function descSorter(a, b)
	return a > b
end

-- allows to iterate a table based on key alphabetical order
local function pairsByKey (t, f)
	local a = {}
	for n in pairs(t) do table.insert(a, n) end
	table.sort(a, f)
	local i = 0      -- iterator variable
	local iter = function ()   -- iterator function
		i = i + 1
		if a[i] == nil then return nil
		else return a[i], t[a[i]]
		end
	end
	return iter
end

--------------------------------------------------------------------------------
-- Standard Addon methods

function AS:OnInitialize()

end

function AS:OnEnable()
	-- Setting up
    AsmoRaidTracker.synchronizer = self
	self:SetOfficers(AsmoRaidTracker.db.profile.officers)
	
	-- Register comms
    self:RegisterComm("ART_Synchronizer", "ListenerOnMessage")
    
    -- Register log on event (guild roster update to ensure that guild roster info is accurate) 
    self:RegisterEvent("GUILD_ROSTER_UPDATE", "OnEnteringWorld")
    -- Schedule a repeating timer to check for translations
    self:ScheduleRepeatingTimer("CheckTranslations", 60)
end

function AS:OnDisable()
    -- Called when the addon is disabled

end

--------------------------------------------------------------------------------
-- Events handling
function AS:OnEnteringWorld()
	-- unregister event after first
	self:UnregisterEvent("GUILD_ROSTER_UPDATE")
	
	-- check if raid ongoing to filter zoning events
	if AsmoRaidTracker.currentRaid then return end
	
	-- If no raid ongoing, broadcast for any raid that could be ongoing
	self:BroadcastForExistingRaid()
end

function AS:CheckTranslations()
	local items = AsmoRaidTracker.db.profile.items
	
	local untranslatedItems = {}
	local translationUpdated = false
	for itemId, item in pairs(items) do
		if itemId ~= "length" and item.locale ~= AsmoRaidTracker.db.profile.language then
			-- Attempt to translate locally if possible
			if GetLocale() == AsmoRaidTracker.db.profile.language then
				local itemName = GetItemInfo(item.itemId)
				if itemName then
					items[item.itemId].itemName = itemName
					items[item.itemId].locale = GetLocale()
					translationUpdated = true
				else
					tinsert(untranslatedItems, item.itemId)
				end
			else
				tinsert(untranslatedItems, item.itemId)
			end
		end
	end
	
	if #untranslatedItems > 0 then
		self:BroadcastForTranslation(untranslatedItems)
	end
	
	-- Update window
	if translationUpdated then
		local raidKey = AsmoRaidTracker.displayState.selectedRaid
		AsmoRaidTracker:UpdateDisplay_LootList(raidKey)
	end
end
--------------------------------------------------------------------------------
-- Methods

function AS:SetOfficers(officers)
	self.officers = {}
	for _, officer in pairs(officers) do
		self.officers[officer] = true
	end
end

function AS:CheckOnline(targets)
	GuildRoster()
	local index = 1
	local name, _, _, _, _, _, _, _, online = GetGuildRosterInfo(index)
	while name do
		if not online or not targets[name] or name == UnitName("player") then
			targets[name] = nil
		else
		end
		index = index + 1
		name, _, _, _, _, _, _, _, online = GetGuildRosterInfo(index)
	end
end

function AS:AddSlave(slaveName)
	local index = #(self.slaveNames)
	index = index + 1
	self.slaveNames[index] = slaveName
	return index
end

--------------------------------------------------------------------------------
-- sync RFC generic method

function AS:DispatchCall(method, ...)
	if AS[method] then
		self[method](self, ...)
		return true
	else
		return false, fmt("Custom error : no listener found for action %s", action)
	end
end

function AS:ListenerOnMessage(prefix, msg, distribution, from)
	-- Generate CommLog entry
	local comm = {
	    type = "IN",
		actor = from,
	}
	
	-- Decode message
	msg = libZip:GetAddonEncodeTable():Decode(msg)
	
	-- Decompress message
	local errMsg
	msg, errMsg = libZip:Decompress(msg)
	
	if not msg then
		self:Print("Error decompressing : ", errMsg)
		comm.error="Decompression error : "..errMsg	
	end
	
	-- Deserialize message
	local noError, method, args = self:Deserialize(msg)
	if not noError then
	    comm.error="Deserialize error : "..action
	else
	    comm.method = method
	    comm.args = ttrunc(args)
	end
	
	-- Call registered listener
	if noError then
		local noError, errorMsg = self:DispatchCall(method, unpack(args))
		if not noError then
			comm.error = errorMsg
		end
	end
	
	-- Log message
	AsmoRaidTracker:DbLogComm(comm)
end

local function reportProgress(target, sent, total)
	local self = AsmoRaidTracker
	if self.displayState.isWindowOpened then
		self.displayWidgets.window:SetStatusText(fmt("Message progress : @%s %.1f%%", target, sent/total*100))
	end
end

-- call Method method on receivers clients with args ...
-- receivers can be either a [targetName]=true list, a 1..n indexed table with names in it or a target name
function AS:SendMessage(receivers, method, ...)
	local targets
	if type(receivers) == "string" then
		targets = {
			[receivers] = true
		}
	elseif type(receivers) == "table" then
		targets = {}
		if receivers[1] then -- assume it is a 1..n table
			for _, target in pairs(receivers) do
				targets[target] = true
			end
		else
			targets = copyvalue(receivers)
		end
	else
		self:Print("Weird case : SendMessage targets table is not a string nor a table", receivers)
		return
	end
	
	if not self.forceNextMessageSend then
		self:CheckOnline(targets)
	else
		self.forceNextMessageSend = false
	end
	
	-- Generate CommLog entry
	local actors = {}
	for target, _ in pairs(targets) do
		tinsert(actors, target)
	end
	
	local comm = {
	    type = "OUT",
		actor = tconcat(actors,","),
		method = method,
	}
	
	-- generate args table
	local args = {}
	for i = 1, select("#", ...) do
	    args[i] = select(i, ...)
	end
	comm.args = ttrunc(args)
	
	-- serialization
	local msg = self:Serialize(method, args)
		
	-- compress message
	msg = libZip:CompressLZW(msg)
	
	-- encode message
	msg = libZip:GetAddonEncodeTable():Encode(msg)
	
	-- Call registered listener
	for target, _ in pairs(targets) do
		self:SendCommMessage("ART_Synchronizer", msg, "WHISPER", target, "NORMAL", reportProgress, target)
	end
	
	-- Log message
	AsmoRaidTracker:DbLogComm(comm)
end

function AS:CancelTimeoutTimer() 
	if self.masterTimeoutTimer then 
		self:CancelTimer(self.masterTimeoutTimer, true)
		self.masterTimeoutTimer = nil
	end
end

function AS:CancelSynchronizationTimer()
	if self.synchronizationTimer then 
		self:CancelTimer(self.synchronizationTimer, true)
		self.synchronizationTimer = nil
	end
end

--------------------------------------------------------------------------------
-- Config synchronisation

function AS:BroadcastFullSynchro(targetName)
	local db = AsmoRaidTracker.db.profile
	self:SendMessage(targetName, "SetFullSynchro", db.raids, db.items, db.itemBlacklist, db.itemRarityThreshold, db.evalDkpValueString, db.language, db.officers, db.raiders, db.excludedGuildRanks)
end
function AS:SetFullSynchro(raids, items, itemBlacklist, itemRarityThreshold, evalDkpValueString, language, officers, raiders, excludedGuildRanks)
	local db = AsmoRaidTracker.db.profile
	db.raids = raids
	db.items = items
	db.itemBlacklist = itemBlacklist
	db.itemRarityThreshold = itemRarityThreshold
	db.evalDkpValueString = evalDkpValueString
	db.language = language
	db.officers = officers
	self:SetOfficers(officers)
	db.raiders = raiders
	db.excludedGuildRanks = excludedGuildRanks
	
	-- Compute stats for inviter
	AsmoRaidTracker.inviter:ComputeStats()
	
	self:Print("Full synchro reçue, t'es opé gars !")
end

function AS:SendConfig(config)
	self:SendMessage(self.officers, "SetConfig", config)
end
function AS:SetConfig(config)
	local db = AsmoRaidTracker.db.profile
	db.itemRarityThreshold = config.itemRarityThreshold
	db.language = config.language
	db.excludedGuildRanks = config.excludedGuildRanks
	db.officers = config.officers
	db.itemBlacklist = config.itemBlacklist
	self:SetOfficers(db.officers)	
end

--------------------------------------------------------------------------------
-- Raid beginning handling

-- Broadcast raid begin
function AS:BroadcastRaidBegin(masterName, raid, raidersStatus)
	-- Cancel timeout and synchronization timers if exists
	self:CancelTimeoutTimer()
	self:CancelSynchronizationTimer()
	-- Send message
	self:SendMessage(self.officers, "BeginRaid", masterName, raid, raidersStatus)
	self.masterName = nil
	self.slaveNames = {}
	self.slaveRank = nil
	
	-- Schedule synchronization timer
	self.synchronizationTimer = self:ScheduleRepeatingTimer("BroadcastSynchronization", BROADCASTRAIDUPDATE_DELAY)
end
function AS:BeginRaid(masterName, raid, raidersStatus)
	-- Cancel timeout and synchronization timers if exists
	self:CancelTimeoutTimer()
	self:CancelSynchronizationTimer()
	-- Init raid
	AsmoRaidTracker:Remote_InitRaid(raid, raidersStatus)
	-- Init synchronizer client/server attributes
	self.masterName = masterName
	self.slaveNames = {}
	self.slaveRank = 1
	self:AcknowledgeRaidMaster(masterName)
	-- Update window
	AsmoRaidTracker:UpdateDisplay_SelectedRaid("ASMORAIDTRACKER_RAIDBEGIN")
end

-- Acknowledge raid master
function AS:AcknowledgeRaidMaster(masterName)
	self.forceNextMessageSend = true -- We know master is online, skip the next message send checkOnline call
	self:SendMessage(masterName, "RegisterSlave", UnitName("player"))
end
function AS:RegisterSlave(slaveName)
	local index = self:AddSlave(slaveName)
	self:AttributeSlaveRank(slaveName, index)
end

-- Attribute rank for slave
function AS:AttributeSlaveRank(slaveName, index)
	self.forceNextMessageSend = true -- We know slave is online, skip the next message send checkOnline call
	self:SendMessage(slaveName, "SetRank", index)
end
function AS:SetRank(index)
	self.slaveRank = index
	self.masterTimeoutTimer = self:ScheduleTimer("MasterTimedOut_AssumeControl", BROADCASTRAIDUPDATE_DELAY + self.slaveRank*BROADCASTRAIDUPDATE_TIMEOUT)
end

--------------------------------------------------------------------------------
-- Raid ending handling

-- Broadcast raid end
function AS:BroadcastRaidEnd(timestamp)
	self:SendMessage(self.slaveNames, "EndRaid", timestamp)
	self.masterName = nil
	self.slaveNames = nil
	self.slaveRank = nil
	
	-- Cancel synchronization timer
	self:CancelSynchronizationTimer()
end
function AS:EndRaid(timestamp)
	AsmoRaidTracker:Remote_EndRaid(timestamp)
	self.masterName = nil
	self.slaveNames = nil
	self.slaveRank = nil
	
	-- Cancel time out timer
	self:CancelTimeoutTimer()
	-- Update window
	AsmoRaidTracker:UpdateDisplay_SelectedRaid("ASMORAIDTRACKER_REMOTERAIDEND")
end

--------------------------------------------------------------------------------
-- Officer logging in handling

-- Broadcasting for existing raid
function AS:BroadcastForExistingRaid()
	self:SendMessage(self.officers, "GetOngoingRaid", UnitName("player"))
end
function AS:GetOngoingRaid(officerName)
	local raid = AsmoRaidTracker.currentRaid
	local raidersStatus = AsmoRaidTracker.currentRaid_MembersStatus
	-- If there is no ongoing raid, or there is but user is not master then do nothing
	if not raid or self.masterName then	return end
	
	-- Retrieve rank if exist
	local rank
	for index, slaveName in pairs(self.slaveNames) do
		if slaveName == officerName then
			rank = index
			break
		end
	end
	-- if officer not in slave list he must be added
	if not rank then
		rank = self:AddSlave(slaveName)
	end
	
	-- Send info to slave
	self:SendOngoingRaidToSlave(officerName, raid, raidersStatus, rank)  	
end

-- Send the ongoing raid to slave
function AS:SendOngoingRaidToSlave(slaveName, raid, raidersStatus, rank)
	self.forceNextMessageSend = true -- We know slave is online, skip the next message send checkOnline call
	self:SendMessage(slaveName, "SetOngoingRaid", raid, raidersStatus, rank, UnitName("player"))
end
function AS:SetOngoingRaid(raid, raidersStatus, rank, masterName)
	AsmoRaidTracker:Remote_InitRaid(raid, raidersStatus)
	self.masterName = masterName
	self.slaveNames = nil
	self.slaveRank = rank
	
	-- Update window
	AsmoRaidTracker:UpdateDisplay_SelectedRaid("ASMORAIDTRACKER_REMOTERAIDBEGIN")
end

--------------------------------------------------------------------------------
-- Master logging out handling
function AS:MasterTimedOut_AssumeControl()
	local raid = AsmoRaidTracker.currentRaid
	local raidersStatus = AsmoRaidTracker.currentRaid_MembersStatus
	
	self:BroadcastRaidBegin(UnitName("player"), raid, raidersStatus)
	
	-- Update window
	AsmoRaidTracker:UpdateDisplay_SelectedRaid("ASMORAIDTRACKER_ASSUMECONTROL_MASTERTIMEDOUT")
end

--------------------------------------------------------------------------------
-- Synchronization handling

-- Broadcasting config
function AS:BroadcastSynchronization()
	local raid, raidersStatus
	
	if self.needSync then
		raid = AsmoRaidTracker.currentRaid
		raidersStatus = AsmoRaidTracker.currentRaid_MembersStatus
		self.needSync = false
	end
	
	-- send message to slaves
	self:SendMessage(self.slaveNames, "Synchronize", raid, raidersStatus)
end
function AS:Synchronize(raid, raidersStatus)
	self:CancelTimeoutTimer()
	if not self.slaveRank then return end
	self.masterTimeoutTimer = self:ScheduleTimer("MasterTimedOut_AssumeControl", BROADCASTRAIDUPDATE_DELAY + self.slaveRank*BROADCASTRAIDUPDATE_TIMEOUT)
	if raid then
		AsmoRaidTracker.db.profile.raids[raid.beginningTimestamp] = raid
		AsmoRaidTracker.currentRaid = raid
		AsmoRaidTracker.currentRaid_MembersStatus = raidersStatus
		-- Update window
		AsmoRaidTracker:UpdateDisplay_RaidRoster(raid.beginningTimestamp)
		AsmoRaidTracker:UpdateDisplay_LootList(raid.beginningTimestamp)
	end
end

--------------------------------------------------------------------------------
-- Loot broadcasting
function AS:NotifyLootToMaster(loot, item)
	-- If user is master don't broadcast
	if not self.masterName then return end
	self:SendMessage(self.masterName, "LootSpotted", loot, item)
end
function AS:LootSpotted(remoteLoot, remoteItem)
	local lootFound = false
	local timestamp = time()
	local raid = AsmoRaidTracker.currentRaid
	for _, localLoot in pairsByKey(raid.loots, descSorter) do
		-- If loot has same id and same original looter then chances are it is the same loot
		if 
			remoteLoot.itemId == localLoot.itemId and
			remoteLoot.looterName == localLoot.originalLooterName
		then
			lootFound = true
			break		
		end
	end
	-- If loot not found
	if not lootFound then
		remoteLoot.timestamp = timestamp
		AsmoRaidTracker:DbAddLoot(raid.beginningTimestamp, timestamp, remoteLoot.looterName, remoteLoot.itemId, remoteItem.itemName, remoteItem.itemTexture, remoteItem.itemRarity, remoteItem.itemLevel, remoteLoot.dkpValue, remoteItem.locale)
	-- If loot is found it may not be saved in the right locale, in which case we update the name
	else
		local localItem = AsmoRaidTracker.db.profile.items[remoteLoot.itemId]
		if localItem.locale ~= AsmoRaidTracker.db.profile.language and remoteItem.locale == AsmoRaidTracker.db.profile.language then
			localItem.itemName = remoteItem.itemName
			localItem.locale = remoteItem.locale
		end  
	end
	
	-- Update window
	AsmoRaidTracker:UpdateDisplay_LootList(raid.beginningTimestamp)
	
	self.needSync = true
end

--------------------------------------------------------------------------------
-- Item definition broadcasting
function AS:BroadcastForItem(itemId)
	self:SendMessage(self.officers, "GetItem", itemId, UnitName("player"))
end
function AS:GetItem(itemId, player)
	local item = AsmoRaidTracker.db.profile.items[itemId]
	if not item then return end
	self.forceNextMessageSend = true -- We know player is online, skip the next message send checkOnline call
	self:SendMessage(player, "SetItem", item)
end
function AS:SetItem(remoteItem)
	
	if not AsmoRaidTracker.db.profile.items[remoteItem.itemId] then
		AsmoRaidTracker:DbAddItem(remoteItem.itemId, remoteItem.itemName, remoteItem.itemTexture, remoteItem.itemRarity, remoteItem.itemLevel, remoteItem.locale)
	end
	-- Loot may not be saved in the right locale, if sent item is of the right locale we update name
	local localItem = AsmoRaidTracker.db.profile.items[remoteItem.itemId]
	if localItem.locale ~= AsmoRaidTracker.db.profile.language and remoteItem.locale == AsmoRaidTracker.db.profile.language then
		localItem.itemName = remoteItem.itemName
		localItem.locale = remoteItem.locale
	end
		
	-- Update window
	local raidKey = AsmoRaidTracker.displayState.selectedRaid
	AsmoRaidTracker:UpdateDisplay_LootList(raidKey)
end

--------------------------------------------------------------------------------
-- Item translation broadcasting
function AS:BroadcastForTranslation(untranslatedItems)
	self:SendMessage(self.officers, "GetTranslation", untranslatedItems, AsmoRaidTracker.db.profile.language, UnitName("player"))
end
function AS:GetTranslation(untranslatedItems, language, player)
	-- If locale is not the one requested for abort
	if language ~= GetLocale() then return end
	
	local translations = {}
	local foundAtLeastOneTranslation = false
	-- check for translation if available
	for _, itemId in pairs(untranslatedItems) do
		local itemName = GetItemInfo(itemId)
		if itemName then
			translations[itemId] = itemName
			foundAtLeastOneTranslation = true
		end
	end
	
	-- Send translations back to requester if any found
	if foundAtLeastOneTranslation then
		self.forceNextMessageSend = true -- We know player is online, skip the next message send checkOnline call
		self:SendMessage(player, "SetTranslation", translations, language)
	end
end
function AS:SetTranslation(translations, language)
	local items = AsmoRaidTracker.db.profile.items
	
	for itemId, itemName in pairs(translations) do
		items[itemId].itemName = itemName
		items[itemId].locale = language
	end
	
	-- Update window
	local raidKey = AsmoRaidTracker.displayState.selectedRaid
	AsmoRaidTracker:UpdateDisplay_LootList(raidKey)
end

--------------------------------------------------------------------------------
-- Raid roster broadcasting
function AS:BroadcastRaidRoster()
	self:SendMessage(self.officers, "SetRaidRoster", AsmoRaidTracker.db.profile.raiders)
end
function AS:SetRaidRoster(raiders)
    AsmoRaidTracker.db.profile.raiders = raiders
end

--------------------------------------------------------------------------------
-- Raid broadcasting
function AS:BroadcastRaid(raid, items)
	self:SendMessage(self.officers, "SetRaid", raid, items)
end
function AS:SetRaid(raid, remoteItems)
    AsmoRaidTracker.db.profile.raids[raid.beginningTimestamp] = raid
    local localItems = AsmoRaidTracker.db.profile.items
    for _, remoteItem in pairs(remoteItems) do
    	local localItem = localItems[remoteItem.itemId]
    	if not localItem or remoteItem.locale == AsmoRaidTracker.db.profile.language then
    		localItems[remoteItem.itemId] = remoteItem
    	end
    end
end